from abc import ABC
from dataclasses import dataclass
from typing import List, Iterable

from oaklib.interfaces.basic_ontology_interface import BasicOntologyInterface
from oaklib.types import CURIE, LABEL, URI, PRED_CURIE


@dataclass
class SearchConfiguration:
    """
    Parameters for altering behavior of search

    .. note ::

        many of these parameters are not yet implemented
    """
    search_terms: List[str] = None
    include_id: bool = True
    include_label: bool = True
    include_aliases: bool = True
    include_definition: bool = False
    complete: bool = False
    is_regex: bool = False

    def use_label_only(self) -> "SearchConfiguration":
        self.include_label = False
        self.include_id = False
        self.include_definition = False
        self.include_aliases = False
        return self


class SearchInterface(BasicOntologyInterface, ABC):
    """
    An Ontology Interface that supports text-based search
    """

    def basic_search(self, search_term: str, config: SearchConfiguration = None) -> Iterable[CURIE]:
        """
        Search over ontology using the specified search term.

        The search term may be a CURIE, label, or alias

        :param search_term:
        :param config:
        :return:
        """
        raise NotImplementedError

    def multiterm_search(self, search_terms: List[str], config: SearchConfiguration = None) -> Iterable[CURIE]:
        """
        As basic_search, using multiple terms

        :param search_terms:
        :param config:
        :return:
        """
        seen = set()
        for t in search_terms:
            if ':' in t:
                yield t
            for curie in self.basic_search(t):
                if curie in seen:
                    continue
                seen.add(curie)
                yield curie






import logging
from abc import ABC
from collections import defaultdict
from dataclasses import dataclass
from typing import List, Any, Iterable, Optional, Type, Dict, Union, Tuple, Iterator

from linkml_runtime import SchemaView
from linkml_runtime.utils.introspection import package_schemaview
from linkml_runtime.utils.metamodelcore import URIorCURIE
from oaklib.implementations.sqldb.model import Statements, Edge, HasSynonymStatement, \
    HasTextDefinitionStatement, ClassNode, IriNode, RdfsLabelStatement, DeprecatedNode, EntailedEdge, \
    ObjectPropertyNode, AnnotationPropertyNode, NamedIndividualNode, HasMappingStatement
from oaklib.interfaces import SubsetterInterface
from oaklib.interfaces.basic_ontology_interface import RELATIONSHIP_MAP, PRED_CURIE, ALIAS_MAP, RELATIONSHIP
from oaklib.interfaces.mapping_provider_interface import MappingProviderInterface
from oaklib.interfaces.obograph_interface import OboGraphInterface
from oaklib.interfaces.relation_graph_interface import RelationGraphInterface
from oaklib.interfaces.search_interface import SearchInterface, SearchConfiguration
from oaklib.interfaces.validator_interface import ValidatorInterface
from oaklib.types import CURIE, SUBSET_CURIE
from oaklib.datamodels import obograph, ontology_metadata
import oaklib.datamodels.validation_datamodel as vdm
from oaklib.datamodels.vocabulary import SYNONYM_PREDICATES, omd_slots, LABEL_PREDICATE, IN_SUBSET
from oaklib.utilities.graph.networkx_bridge import transitive_reduction_by_predicate
from sqlalchemy import select, text, exists
from sqlalchemy.orm import sessionmaker, aliased
from sqlalchemy import create_engine


# TODO: move to schemaview
def get_range_xsd_type(sv: SchemaView, rng: str) -> Optional[URIorCURIE]:
    t = sv.get_type(rng)
    if t.uri:
        return t.uri
    elif t.typeof:
        return get_range_xsd_type(sv, t.typeof)
    else:
        raise ValueError(f'No xsd type for {rng}')



@dataclass
class SqlImplementation(RelationGraphInterface, OboGraphInterface, ValidatorInterface, SearchInterface,
                        SubsetterInterface, MappingProviderInterface, ABC):
    """
    A :class:`OntologyInterface` implementation that wraps a SQL Relational Database

    This could be a local file (accessed via SQL Lite) or a local/remote server (e.g PostgreSQL)

    The schema is assumed to follow the `semantic-sql <https://github.com/incatools/semantic-sql>`_ schema

    This uses SQLAlchemy ORM Models:

    - :class:`Statements`
    - :class:`Edge`
    """
    # TODO: use SQLA types
    engine: Any = None
    _session: Any = None
    _connection: Any = None
    _ontology_metadata_model: SchemaView = None

    def __post_init__(self):
        if self.engine is None:
            self.engine = create_engine(self.resource.slug)  ## TODO

    @property
    def session(self):
        if self._session is None:
            session_cls = sessionmaker(self.engine)
            self._session = session_cls()
        return self._session

    @property
    def connection(self):
        if self._connection is None:
            self._connection = self.engine.connect()
        return self._session

    @property
    def ontology_metadata_model(self):
        if self._ontology_metadata_model is None:
            self._ontology_metadata_model = package_schemaview(ontology_metadata.__name__)
        return self._ontology_metadata_model

    def all_entity_curies(self) -> Iterable[CURIE]:
        s = text('SELECT id FROM class_node WHERE id NOT LIKE "\_:%" ESCAPE "\\"')
        for row in self.engine.execute(s):
            yield row['id']

    def get_label_by_curie(self, curie: CURIE) -> Optional[str]:
        s = text('SELECT value FROM rdfs_label_statement WHERE subject = :curie')
        for row in self.engine.execute(s, curie=curie):
            return row['value']

    def get_labels_for_curies(self, curies: Iterable[CURIE]) -> Iterable[Tuple[CURIE, str]]:
        for row in self.session.query(RdfsLabelStatement).filter(RdfsLabelStatement.subject.in_(tuple(list(curies)))):
            yield row.subject, row.value

    def alias_map_by_curie(self, curie: CURIE) -> ALIAS_MAP:
        m = defaultdict(list)
        m[LABEL_PREDICATE] = [self.get_label_by_curie(curie)]
        for row in self.session.query(HasSynonymStatement).filter(HasSynonymStatement.subject == curie):
            m[row.predicate].append(row.value)
        return m

    def _get_subset_curie(self, curie: str) -> str:
        if '#' in curie:
            return curie.split('#')[-1]
        else:
            return curie

    def _subset_uri_to_curie_map(self) -> Dict[str, CURIE]:
        m = {}
        for row in self.session.query(Statements.object).filter(Statements.predicate == IN_SUBSET):
            uri = row.object
            m[uri] = self._get_subset_curie(row.object)
        return m

    def _subset_curie_to_uri_map(self) -> Dict[CURIE, str]:
        m = {}
        for row in self.session.query(Statements.object).filter(Statements.predicate == IN_SUBSET):
            uri = row.object
            m[self._get_subset_curie(row.object)] = uri
        return m

    def all_subset_curies(self) -> Iterable[SUBSET_CURIE]:
        for s in self._subset_curie_to_uri_map().keys():
            yield s

    def curies_by_subset(self, subset: SUBSET_CURIE) -> Iterable[CURIE]:
        sm = self._subset_curie_to_uri_map()
        for row in self.session.query(Statements.subject).filter(Statements.predicate == IN_SUBSET,
                                                                 Statements.object == sm[subset]):
            yield self._get_subset_curie(row.subject)

    def basic_search(self, search_term: str, config: SearchConfiguration = SearchConfiguration()) -> Iterable[CURIE]:
        search_term = f'%{search_term}%'
        preds = []
        if config.include_label:
            preds.append(omd_slots.label.curie)
        if config.include_aliases:
            preds += SYNONYM_PREDICATES
        view = Statements
        q = self.session.query(view.subject).filter(view.predicate.in_(tuple(preds))).filter(
            view.value.like(search_term))
        for row in q.distinct():
            yield str(row.subject)

    def get_outgoing_relationships_by_curie(self, curie: CURIE, isa_only: bool = False) -> RELATIONSHIP_MAP:
        rmap = defaultdict(list)
        for row in self.session.query(Edge).filter(Edge.subject == curie):
            rmap[row.predicate].append(row.object)
        return rmap

    def get_incoming_relationships_by_curie(self, curie: CURIE) -> RELATIONSHIP_MAP:
        rmap = defaultdict(list)
        for row in self.session.query(Edge).filter(Edge.object == curie):
            rmap[row.predicate].append(row.subject)
        return rmap

    def get_simple_mappings_by_curie(self, curie: CURIE) -> RELATIONSHIP_MAP:
        m = defaultdict(list)
        for row in self.session.query(HasMappingStatement).filter(HasMappingStatement.subject == curie):
            m[row.predicate].append(row.value)
        return m

    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    # Implements: OboGraphInterface
    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    def node(self, curie: CURIE) -> obograph.Node:
        meta = obograph.Meta()
        n = obograph.Node(id=curie, meta=meta)
        for row in self.session.query(Statements).filter(Statements.subject == curie):
            if row.value is not None:
                v = row.value
            elif row.object is not None:
                v = row.object
            else:
                continue
            pred = row.predicate
            if pred == omd_slots.label.curie:
                n.label = v
            else:
                if pred == omd_slots.definition.curie:
                    meta.definition = obograph.DefinitionPropertyValue(val=v)
        return n

    def ancestors(self, start_curies: Union[CURIE, List[CURIE]], predicates: List[PRED_CURIE] = None) -> Iterable[
        CURIE]:
        q = self.session.query(EntailedEdge)
        if isinstance(start_curies, list):
            q = q.filter(EntailedEdge.subject.in_(tuple(start_curies)))
        else:
            q = q.filter(EntailedEdge.subject == start_curies)
        if predicates is not None:
            q = q.filter(EntailedEdge.predicate.in_(tuple(predicates)))
        logging.debug(f'Ancestors query: {q}')
        for row in q:
            yield row.object

    def descendants(self, start_curies: Union[CURIE, List[CURIE]], predicates: List[PRED_CURIE] = None) -> Iterable[
        CURIE]:
        q = self.session.query(EntailedEdge)
        if isinstance(start_curies, list):
            q = q.filter(EntailedEdge.object.in_(tuple(start_curies)))
        else:
            q = q.filter(EntailedEdge.object == start_curies)
        if predicates is not None:
            q = q.filter(EntailedEdge.predicate.in_(tuple(predicates)))
        for row in q:
            yield row.subject

    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    # Implements: ValidatorInterface
    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    def validate(self, configuration: vdm.ValidationConfiguration = None) -> Iterable[vdm.ValidationResult]:
        if configuration and configuration.schema_path:
            sv = SchemaView(configuration.schema_path)
            self._ontology_metadata_model = sv
        else:
            sv = self.ontology_metadata_model
        for slot_name in sv.all_slots():
            for r in self._check_slot(slot_name):
                yield r
        for r in self._check_for_unknown_slots():
            yield r

    def _missing_value(self, predicate_table: Type, type_table: Type = ClassNode) -> Iterable[CURIE]:
        pred_subq = self.session.query(predicate_table.subject)
        obs_subq = self.session.query(DeprecatedNode.id)
        main_q = self.session.query(type_table).join(IriNode, type_table.id == IriNode.id)
        for row in main_q.filter(type_table.id.not_in(pred_subq)).filter(type_table.id.not_in(obs_subq)):
            yield row.id

    def term_curies_without_definitions(self) -> Iterable[CURIE]:
        return self._missing_value(HasTextDefinitionStatement)

    def term_curies_without_labels(self) -> Iterable[CURIE]:
        return self._missing_value(RdfsLabelStatement)

    def _check_for_unknown_slots(self) -> Iterable[vdm.ValidationResult]:
        sv = self.ontology_metadata_model
        preds = [sv.get_uri(s, expand=False) for s in sv.all_slots().values()]
        logging.info(f'Known preds: {len(preds)} -- checking for other uses')
        main_q = self.session.query(Statements).filter(Statements.predicate.not_in(preds)).join(IriNode, Statements.subject == IriNode.id)
        try:
            for row in main_q:
                result = vdm.ValidationResult(subject=row.subject,
                                              predicate=row.predicate,
                                              severity=vdm.SeverityOptions.ERROR,
                                              type=vdm.ValidationResultType.ClosedConstraintComponent.meaning,
                                              info=f'Unknown pred ({row.predicate}) = {row.object} {row.value}'
                                              )
                yield result
        except ValueError as e:
            logging.error(f'EXCEPTION: {e}')
            pass

    def _check_slot(self, slot_name: str, class_name: str = 'Class') -> Iterable[vdm.ValidationResult]:
        """
        Validates all data with respect to a specific slot

        :param slot_name:
        :param class_name:
        :return:
        """
        sv = self.ontology_metadata_model
        class_cls = sv.get_class(class_name)
        # for efficiency we map directly to table/view names rather
        # than querying over rdf:type; this allows for optimization via view materialization
        if class_name == 'Class':
            sqla_cls = ClassNode
        elif class_name == 'ObjectProperty':
            sqla_cls = ObjectPropertyNode
        elif class_name == 'AnnotationProperty':
            sqla_cls = AnnotationPropertyNode
        elif class_name == 'NamedIndividual':
            sqla_cls = NamedIndividualNode
        else:
            raise NotImplementedError(f'cannot handle {class_name}')
        slot = sv.induced_slot(slot_name, class_name)
        if slot.designates_type:
            logging.info(f'Ignoring type designator: {slot_name}')
            return
        logging.info(f'Validating: {slot_name}')
        predicate = sv.get_uri(slot, expand=False)
        is_used = self.session.query(Statements.predicate).filter(Statements.predicate == predicate).first() is not None
        pred_subq = self.session.query(Statements.subject).filter(Statements.predicate == predicate)
        obs_subq = self.session.query(DeprecatedNode.id)
        if (slot.required or slot.recommended) and not slot.identifier:
            # MinCardinality == 1
            if slot.required:
                severity = vdm.SeverityOptions.ERROR
            else:
                severity = vdm.SeverityOptions.WARNING
            logging.info(f'MinCard check: Leaving off: {slot_name} is {severity.text}')
            # exclude blank nodes
            main_q = self.session.query(sqla_cls).join(IriNode, sqla_cls.id == IriNode.id)
            main_q = main_q.filter(sqla_cls.id.not_in(pred_subq))
            main_q = main_q.filter(sqla_cls.id.not_in(obs_subq))
            for row in main_q:
                result = vdm.ValidationResult(subject=row.id,
                                              predicate=predicate,
                                              severity=severity,
                                              type=vdm.ValidationResultType.MinCountConstraintComponent.meaning,
                                              info=f'Missing slot ({slot_name}) for {row.id}'
                                              )
                yield result
        if not is_used:
            return
        if slot.deprecated:
            main_q = self.session.query(Statements.subject).filter(Statements.predicate == predicate)
            main_q = main_q.join(sqla_cls, Statements.subject == sqla_cls.id)
            for row in main_q:
                result = vdm.ValidationResult(subject=row.subject,
                                              predicate=predicate,
                                              severity=vdm.SeverityOptions.WARNING,
                                              type=vdm.ValidationResultType.DeprecatedPropertyComponent.meaning,
                                              info=f'Deprecated slot ({slot_name}) for {row.subject}'
                                              )
                yield result
        if not slot.multivalued:
            # MaxCardinality == 1
            # TODO
            is_object_iri = slot.range in sv.all_classes()
            st1 = aliased(Statements)
            st2 = aliased(Statements)
            main_q = self.session.query(st1.subject).join(st2, st1.subject == st2.subject)
            main_q = main_q.filter(st1.predicate == predicate)
            main_q = main_q.filter(st2.predicate == predicate)
            if is_object_iri:
                main_q = main_q.filter(st1.object != st2.object)
            else:
                main_q = main_q.filter(st1.value != st2.value)
            main_q = main_q.join(sqla_cls, st1.subject == sqla_cls.id)
            for row in main_q:
                result = vdm.ValidationResult(subject=row.subject,
                                              predicate=predicate,
                                              severity=vdm.SeverityOptions.ERROR,
                                              type=vdm.ValidationResultType.MaxCountConstraintComponent.meaning,
                                              info=f'Too many vals for {slot_name}'
                                              )
                yield result
        if slot.range:
            rng = slot.range
            rng_elements = sv.slot_applicable_range_elements(slot)
            # for now we don't handle Union or Any
            if len(rng_elements) < 2:
                logging.info(f'Datatype check: {slot_name} range is {rng_elements}')
                is_object_iri = rng in sv.all_classes()
                if is_object_iri:
                    constr = Statements.object.is_(None)
                else:
                    constr = Statements.value.is_(None)
                main_q = self.session.query(Statements.subject)
                main_q = main_q.join(IriNode, Statements.subject == IriNode.id)
                main_q = main_q.join(sqla_cls, Statements.subject == sqla_cls.id)
                main_q = main_q.filter(Statements.predicate == predicate, constr)
                for row in main_q:
                    result = vdm.ValidationResult(subject=row.subject,
                                                  predicate=predicate,
                                                  severity=vdm.SeverityOptions.ERROR,
                                                  type=vdm.ValidationResultType.DatatypeConstraintComponent.meaning,
                                                  info=f'Incorrect object type for {slot_name} range = {rng} should_be_iri = {is_object_iri}'
                                                  )
                    yield result
                if rng in sv.all_types():
                    uri = get_range_xsd_type(sv, rng)
                    #uri = rng_type.uri
                    main_q = self.session.query(Statements.subject)
                    main_q = main_q.join(IriNode, Statements.subject == IriNode.id)
                    main_q = main_q.join(sqla_cls, Statements.subject == sqla_cls.id)
                    main_q = main_q.filter(Statements.predicate == predicate, Statements.datatype != uri)
                    #print(main_q)
                    for row in main_q:
                        result = vdm.ValidationResult(subject=row.subject,
                                                      predicate=predicate,
                                                      severity=vdm.SeverityOptions.ERROR,
                                                      type=vdm.ValidationResultType.DatatypeConstraintComponent.meaning,
                                                      info=f'Incorrect datatype for {slot_name} expected: {uri} for {rng}'
                                                      )
                        yield result

    def gap_fill_relationships(self, seed_curies: List[CURIE], predicates: List[PRED_CURIE] = None) -> Iterator[RELATIONSHIP]:
        seed_curies = tuple(seed_curies)
        q = self.session.query(EntailedEdge).filter(EntailedEdge.subject.in_(seed_curies))
        q = q.filter(EntailedEdge.object.in_(seed_curies))
        if predicates:
            q = q.filter(EntailedEdge.predicate.in_(tuple(predicates)))
        rels = []
        for row in q:
            if row.subject != row.object:
                rels.append((row.subject, row.predicate, row.object))
        for rel in transitive_reduction_by_predicate(rels):
            yield rel



from abc import ABC
from dataclasses import dataclass
from typing import Optional

from oaklib.implementations.sparql.sparql_implementation import SparqlImplementation
from oaklib.interfaces.mapping_provider_interface import MappingProviderInterface
from oaklib.interfaces.search_interface import SearchInterface

ONTOBEE_MERGED_GRAPH_PREFIX = 'http://purl.obolibrary.org/obo/merged/'

@dataclass
class OntobeeImplementation(SparqlImplementation, SearchInterface, MappingProviderInterface):
    """
    Wraps the Ontobee sparql endpoint

    An OntobeeImplementation can be initialed by:

        .. code:: python

           >>>  oi = OntobeeImplementation.create()

        The default ontobee endpoint will be assumed

    See: `<https://www.ontobee.org/>`_
    """

    def _default_url(self) -> str:
        return "http://sparql.hegroup.org/sparql"

    @property
    def named_graph(self) -> Optional[str]:
        if self.resource.slug is None:
            return None
        else:
            return f'{ONTOBEE_MERGED_GRAPH_PREFIX}{self.resource.slug.upper()}'














import urllib

import requests
import logging
from dataclasses import dataclass, field
from typing import Any, List, Dict, Union, Iterator, Iterable, Tuple

from oaklib.datamodels import oxo
from oaklib.datamodels.oxo import ScopeEnum
from oaklib.datamodels.text_annotator import TextAnnotation
from oaklib.datamodels.vocabulary import IS_A
from oaklib.implementations.ols.oxo_utils import load_oxo_payload
from oaklib.interfaces.basic_ontology_interface import PREFIX_MAP
from oaklib.interfaces.mapping_provider_interface import MappingProviderInterface
from oaklib.interfaces.search_interface import SearchInterface, SearchConfiguration
from oaklib.interfaces.text_annotator_interface import TextAnnotatorInterface
from oaklib.types import CURIE, PRED_CURIE
from sssom import Mapping
from sssom.sssom_datamodel import MatchTypeEnum
from sssom.sssom_document import MappingSetDocument

ANNOTATION = Dict[str, Any]

oxo_pred_mappings = {
    ScopeEnum.EXACT.text: 'skos:exactMatch',
    ScopeEnum.BROADER.text: 'skos:broadMatch',
    ScopeEnum.NARROWER.text: 'skos:narrowMatch',
    ScopeEnum.RELATED.text: 'skos:closeMatch',
}


@dataclass
class OlsImplementation(TextAnnotatorInterface, SearchInterface, MappingProviderInterface):
    """
    Implementation over OLS and OxO APIs

    .. note::


    """
    ols_api_key: str = None
    label_cache: Dict[CURIE, str] = field(default_factory=lambda: {})
    base_url = "https://www.ebi.ac.uk/spot/oxo/api/mappings"
    ols_base_url = "https://www.ebi.ac.uk/ols/api/ontologies/"
    prefix_map: Dict[str, str] = field(default_factory=lambda: {})
    focus_ontology: str = None

    def __post_init__(self):
        if self.focus_ontology is None:
            if self.resource:
                self.focus_ontology = self.resource.slug

    def add_prefix(self, curie: str, uri: str):
        [pfx, local] = curie.split(':', 2)
        if pfx not in self.prefix_map:
            self.prefix_map[pfx] = uri.replace(local, '')

    def get_prefix_map(self) -> PREFIX_MAP:
        return self.prefix_map

    def get_labels_for_curies(self, curies: Iterable[CURIE]) -> Iterable[Tuple[CURIE, str]]:
        raise NotImplementedError

    def annotate_text(self, text: str) -> Iterator[TextAnnotation]:
        raise NotImplementedError

    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    # Implements: OboGraphInterface
    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    def ancestors(self, start_curies: Union[CURIE, List[CURIE]], predicates: List[PRED_CURIE] = None) -> Iterable[CURIE]:
        query = 'hierarchicalAncestors'
        if predicates:
            if predicates == [IS_A]:
                query = 'ancestors'
            elif IS_A not in predicates:
                raise NotImplementedError(f'OLS always include {IS_A}, you selected: {predicates}')
        if not isinstance(start_curies, list):
            start_curies = [start_curies]
        ancs = set()
        ontology = self.focus_ontology
        for curie in start_curies:
            term_id = self.curie_to_uri(curie)
            # must be double encoded https://www.ebi.ac.uk/ols/docs/api
            term_id_quoted = urllib.parse.quote(term_id, safe='')
            term_id_quoted = urllib.parse.quote(term_id_quoted, safe='')
            url = f'{self.ols_base_url}{ontology}/terms/{term_id_quoted}/{query}'
            logging.debug(f'URL={url}')
            result = requests.get(url)
            obj = result.json()
            if result.status_code == 200 and '_embedded' in obj:
                ancs.update([x['obo_id'] for x in obj['_embedded']['terms']])
            else:
                logging.debug(f'No ancestors for {url} (maybe ontology not indexed in OLS?)')
                ancs = []
        return list(ancs)

    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    # Implements: SearchInterface
    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    def basic_search(self, search_term: str, config: SearchConfiguration = SearchConfiguration()) -> Iterable[CURIE]:
        raise NotImplementedError

    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    # Implements: MappingsInterface
    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    def get_sssom_mappings_by_curie(self, curie: Union[str, CURIE]) -> Iterator[Mapping]:
        result = requests.get(self.base_url, params=dict(fromId=curie))
        obj = result.json()
        container = load_oxo_payload(obj)
        return self.convert_payload(container)

    def convert_payload(self, container: oxo.Container) -> Iterator[Mapping]:
        oxo_mappings = container._embedded.mappings
        mappings: Mapping = []
        for oxo_mapping in oxo_mappings:
            oxo_s = oxo_mapping.fromTerm
            oxo_o = oxo_mapping.toTerm
            mapping = Mapping(subject_id=oxo_s.curie,
                              subject_label=oxo_s.label,
                              subject_source=oxo_s.datasource.prefix if oxo_s.datasource else None,
                              predicate_id=oxo_pred_mappings[str(oxo_mapping.scope)],
                              match_type=MatchTypeEnum.Unspecified,
                              object_id=oxo_o.curie,
                              object_label=oxo_o.label,
                              object_source=oxo_o.datasource.prefix if oxo_o.datasource else None,
                              mapping_provider=oxo_mapping.datasource.prefix)
            self.add_prefix(oxo_s.curie, oxo_s.uri)
            self.add_prefix(oxo_o.curie, oxo_o.uri)
            yield mapping

    def fill_gaps(self, msdoc: MappingSetDocument, confidence: float = 1.0) -> int:
        curie_map = curie_to_uri_map(msdoc)
        inv_map = {v: k for k, v in curie_map.items()}
        n = 0
        for curie, uri in curie_map.items():
            pfx, _ = curie.split(':', 2)
            ancs = self.get_ancestors(uri, ontology=pfx.lower())
            logging.debug(f'{curie} ANCS = {ancs}')
            for anc in ancs:
                if anc in curie_map:
                    m = Mapping(subject_id=curie,
                                object_id=anc,
                                predicate_id='rdfs:subClassOf',
                                confidence=confidence,
                                match_type=MatchTypeEnum.HumanCurated
                                )
                    logging.info(f'Gap filled link: {m}')
                    msdoc.mapping_set.mappings.append(m)
                    n += 1
        return n




from abc import ABC
from typing import Dict, List, Tuple, Any, Iterator

from oaklib.interfaces.basic_ontology_interface import BasicOntologyInterface
import rdflib
from oaklib.types import CURIE, PRED_CURIE
from rdflib import URIRef

TRIPLE = Tuple[CURIE, PRED_CURIE, Any]
RDF_TRIPLE = Tuple[URIRef, URIRef, Any]

class RdfInterface(BasicOntologyInterface, ABC):
    """
    presents ontology as simple RDF graphs, using the rdflib datamodel
    """

    def graph(self) -> rdflib.Graph:
        raise NotImplementedError()

    def _triple_as_urirefs(self, triple: TRIPLE) -> RDF_TRIPLE:
        pass

    def _triple_as_curies(self, triple: RDF_TRIPLE) -> TRIPLE:
        pass

    def triples(self, pattern: TRIPLE) -> Iterator[TRIPLE]:
        for t in self.graph().triples(self._triple_as_urirefs(pattern)):
            yield self._triple_as_curies(t)

    def objects(self, subject, predicate) -> Iterator[Any]:
        return self.graph().objects(subject, predicate)

    def extract_triples(self, seed_curies: List[CURIE], predicates: List[PRED_CURIE] = None, strategy: str = None,
                        map_to_curies=True) -> Iterator[TRIPLE]:
        pass




from abc import ABC
from typing import Dict, List, Iterable

from oaklib.interfaces.basic_ontology_interface import BasicOntologyInterface
from oaklib.resource import OntologyResource
from oaklib.types import CURIE


class PatcherInterface(BasicOntologyInterface, ABC):
    """
    Applies diffs

    See `KGCL <https://github.com/cmungall/knowledge-graph-change-language>`_
    """

    def apply_patch(self, patch: str) -> None:
        """
        Applies a change description

        :param patch: TBD use KGCL?
        :return:
        """





from abc import ABC
from dataclasses import dataclass
from typing import Dict, List, Iterable

from oaklib.interfaces.basic_ontology_interface import BasicOntologyInterface
from oaklib.types import CURIE
from oaklib.datamodels.validation_datamodel import ValidationResult, ValidationConfiguration


class ValidatorInterface(BasicOntologyInterface, ABC):
    """
    Basic ontology QC

    The overall goal is to support the following

    - validating against OntologyMetadata schema
    - lexical checks
    - wrapping reasoning
    - structural graph checks

    Specific implementations may choose to implement efficient methods for this.
    For example, a SQL implementation can quickly determine all terms
    missing definitions with a query over an indexed table.

    Currently the main implementation for this is the SqlDatabase implementation,
    this implements a generic property check using the OntologyMetadata datamodel

    See:

     - `OntologyMetadata <https://incatools.github.io/ontology-access-kit/datamodels/ontology-metadata/>`_
    """

    def term_curies_without_definitions(self) -> Iterable[CURIE]:
        """
        TODO: decide whether to write highly specific methods or use a generic validate method
        :return:
        """
        # Implementations are advise to implement more efficient interfaces for their back-end
        for curie in self.all_entity_curies():
            if self.get_definition_by_curie(curie) is None:
                yield curie

    def validate(self, configuration: ValidationConfiguration = None) -> Iterable[ValidationResult]:
        """
        Validate entire ontology or wrapped ontologies

        Validation results might be implementation specific

        - reasoners will yield logical problems
        - shape checkers or schema checkers like SHACL or LinkML will return closed-world structural violations
        - specialized implementations may yield lexical or other kinds of problems
        :return:
        """
        raise NotImplementedError

    def check_external_references(self):
        raise NotImplementedError




from dataclasses import dataclass
from typing import List, Set, Union, Dict, Tuple, Iterable


@dataclass
class ListPair:
    set1_id: str = None
    set2_id: str = None
    list1: List = None
    list2: List = None
    list1_len: int = None
    list2_len: int = None
    intersection_len: int = None
    union_len: int = None
    jaccard_similarity: float = None
    dice_similarity: float = None
    overlap_coefficient: float = None
    subsumes_score: float = None
    subsumed_by_score: float = None



NAMED_COLLECTION = Tuple[str, List]
LIST_OR_SET = Union[List, Set]

def compute_list_pair_statistics(pair: ListPair):
    set1 = set(pair.set1_id)
    set2 = set(pair.set2_id)
    pair.list1_len = len(set1)
    pair.list2_len = len(set2)
    pair.intersection_len = len(set1.intersection(set2))
    pair.union_len = len(set1.union(set2))
    pair.jaccard_similarity = pair.intersection_len / pair.union_len
    pair.dice_similarity = 2 * pair.intersection_len / (pair.list1_len + pair.list2_len)
    pair.overlap_coefficient = pair.intersection_len / min(pair.list1_len, pair.list2_len)
    pair.subsumed_by_score = pair.intersection_len / pair.list1_len
    pair.subsumes_score = pair.intersection_len / pair.list2_len



def compute_all_pairs(sets: Dict[str, List]) -> Iterable[ListPair]:
    """
    compute statistics on all pairs of sets

    :param sets:
    :return:
    """
    for k1, list1 in sets.items():
        for k2, list2 in sets.items():
            r = ListPair(k1, k2)
            compute_list_pair_statistics(r)
            yield r



def setwise_jaccard_similarity(set1: LIST_OR_SET, set2: LIST_OR_SET) -> float:
    set1 = set(set1)
    set2 = set(set2)
    #print(f'T = {len(set1)} / {len(set2)} == {set1} / {set2}')
    return len(set1.intersection(set2)) / len(set1.union(set2))


"""
Utilities for working with networkx
-----------------------------------

NetworkX is a popular python package for working with graphs
"""
try:
    # Python <= 3.9
    from collections import Iterable, defaultdict
except ImportError:
    # Python > 3.9
    from collections.abc import Iterable

import networkx as nx
from oaklib.interfaces.basic_ontology_interface import RELATIONSHIP


def relationships_to_multi_digraph(relationships: Iterable[RELATIONSHIP], reverse: bool = True) -> nx.MultiDiGraph:
    """
    Converts an OBOGraph to NetworkX

    :param relationships:
    :param reverse: treat subject as the networkx parent (default true)
    :return:
    """
    g = nx.MultiDiGraph()
    for rel in relationships:
        if reverse:
            g.add_edge(rel[2], rel[0], predicate=rel[1])
        else:
            g.add_edge(rel[0], rel[2], predicate=rel[1])
    return g

def transitive_reduction(relationships: Iterable[RELATIONSHIP]) -> Iterable[RELATIONSHIP]:
    relationships = list(relationships)
    tuples = [(o, s) for s, p, o in relationships]
    g = nx.DiGraph(tuples)
    reduced = nx.transitive_reduction(g)
    for r in relationships:
        s, p, o = r
        if (o, s) in reduced.edges:
            yield r

def transitive_reduction_by_predicate(relationships: Iterable[RELATIONSHIP]) -> Iterable[RELATIONSHIP]:
    relationships = list(relationships)
    tuples_dict = defaultdict(list)
    rels_dict = defaultdict(list)
    for s, p, o in relationships:
        if o != s:
            tuples_dict[p].append((o, s))
        rels_dict[p].append((s, p, o))
    for p, tuples in tuples_dict.items():
        g = nx.DiGraph(tuples)
        reduced = nx.transitive_reduction(g)
        for r in rels_dict[p]:
            s, _, o = r
            if (o, s) in reduced.edges:
                yield r


from typing import List, Tuple, Dict, Iterable

from oaklib.datamodels.vocabulary import IS_A, PART_OF
from oaklib.interfaces import OntologyInterface, BasicOntologyInterface
from oaklib.interfaces.obograph_interface import OboGraphInterface
from oaklib.types import SUBSET_CURIE, PRED_CURIE, CURIE
from oaklib.utilities.semsim.similarity_utils import setwise_jaccard_similarity, compute_all_pairs, ListPair

SUBSET_DICT = Dict[SUBSET_CURIE, List[CURIE]]
DEFAULT_PREDICATES = [IS_A, PART_OF]


def get_subset_dict(oi: OboGraphInterface) -> SUBSET_DICT:
    """
    Return a dictionary keyed by subset name with value being all subset members

    :param oi:
    :return:
    """
    return {s: list(oi.curies_by_subset(s)) for s in oi.all_subset_curies()}

def terms_by_subsets(oi: OboGraphInterface, remove_empty: bool = True, subsumed_score: float = None, min_subsets: int = None,
                     prefix: str = None) -> Iterable[Tuple]:
    subsets = get_subset_dict(oi)
    subsets = filter_by_prefix(subsets, prefix)
    if remove_empty:
        subsets = {k: v for k, v in subsets.items() if v != []}
    all_curies = set()
    for curies in subsets.values():
        all_curies.update(curies)
    label_map = {curie: label for curie, label in oi.get_labels_for_curies(all_curies)}
    predicates = DEFAULT_PREDICATES
    subset_ancs = {}
    if subsumed_score is not None:
        for subset, subset_curies in subsets.items():
            subset_ancs[subset] = set(oi.ancestors(subset_curies, predicates=predicates))
    for curie in all_curies:
        tups = []
        n = 0
        for subset, subset_curies in subsets.items():
            v = 0.0
            if curie in subset_curies:
                v = 1.0
                n += 1
            else:
                if subsumed_score is not None:
                    #if len(set(oi.ancestors(curie, predicates=predicates)).intersection(subset_curies)) > 0:
                    if subset in subset_ancs and curie in subset_ancs[subset]:
                        v = subsumed_score
            tups.append((curie, label_map[curie], subset, v))
        if min_subsets is not None and n < min_subsets:
            continue
        for tup in tups:
            yield tup


def filter_by_prefix(subsets: SUBSET_DICT, prefix: str) -> SUBSET_DICT:
    if prefix:
        def include(x: str):
            return x.startswith(f'{prefix}:')
        return {k: [x for x in v if include(x)] for k, v in subsets.items()}
    else:
        return subsets


def compare_all_subsets(oi: OboGraphInterface, extend_down: bool = False, remove_empty: bool = True, prefix: str = None) -> Iterable[ListPair]:
    """

    :param oi:
    :return:
    """
    subsets = get_subset_dict(oi)
    if extend_down:
        subsets = extend_subsets_down(oi)
    if prefix:
        def include(x: str):
            return x.startswith(f'{prefix}:')
        subsets = {k: [x for x in v if include(x)] for k, v in subsets.items()}
    if remove_empty:
        subsets = {k: v for k, v in subsets.items() if v != []}
    return compute_all_pairs(subsets)

def extend_subsets_down(oi: OboGraphInterface, predicates: List[PRED_CURIE] = DEFAULT_PREDICATES) -> SUBSET_DICT:
    subsets = get_subset_dict(oi)
    all_curies = set()
    for curies in subsets.values():
        all_curies.update(curies)
    new_subsets = {}
    for subset, curies in subsets.items():
        extended_set = set(oi.descendants(curies, predicates=predicates)).intersection(all_curies)
        new_subsets[subset] = list(extended_set)
    return new_subsets

def subset_overlap(oi: OboGraphInterface, subset1: SUBSET_CURIE, subset2: SUBSET_CURIE, predicates: List[PRED_CURIE] = None) -> float:
    curies1 = list(oi.curies_by_subset(subset1))
    curies2 = list(oi.curies_by_subset(subset2))
    if predicates is None:
        predicates = [IS_A, PART_OF]
    descs1 = oi.descendants(curies1, predicates=predicates)
    descs2 = oi.descendants(curies2, predicates=predicates)
    return setwise_jaccard_similarity(descs1, descs2)

def all_subsets_overlap(oi: OboGraphInterface) -> List[Tuple[float, SUBSET_CURIE, SUBSET_CURIE]]:
    subsets = list(oi.all_subset_curies())
    results = []
    dmap = {s: list(oi.descendants(list(oi.curies_by_subset(s)))) for s in subsets}
    dmap = {k: v for k, v in dmap.items() if len(v) > 0}
    subsets = dmap.keys()
    for s1 in subsets:
        for s2 in subsets:
            if s1 == s2:
                results.append((1.0, s1, s2))
            elif s1 > s2:
                results.append((setwise_jaccard_similarity(dmap[s1], dmap[s2]), s1, s2))
    results.sort(key=lambda tup: 1-tup[0])
    return results




from abc import ABC
from typing import Dict, List

from oaklib.interfaces.basic_ontology_interface import BasicOntologyInterface
from oaklib.types import CURIE, LABEL, URI

# Deprecate? Just use utility class?
class NetworkxInterface(BasicOntologyInterface, ABC):
    """
    TODO: consider direct use of nxontology
    """


import csv
import sys
from dataclasses import dataclass, field
from typing import Any

from linkml_runtime import CurieNamespace
from linkml_runtime.dumpers import yaml_dumper
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from linkml_runtime.utils.yamlutils import YAMLRoot
from oaklib.io.streaming_writer import StreamingWriter


def _keyval(x: Any) -> str:
    if isinstance(x, CurieNamespace):
        return str(x.curie())
    #if isinstance(x, EnumDefinitionImpl):
    #    if x.curie:
    #        return str(x.curie)
    return str(x)


@dataclass
class StreamingYamlWriter(StreamingWriter):
    """
    A writer that emits one document at a time in one stream
    """

    def emit(self, obj: YAMLRoot):
        self.file.write(yaml_dumper.dumps(obj))
        self.file.write("\n---\n")





import logging
from abc import ABC
from collections import defaultdict
from dataclasses import dataclass
from typing import List, Iterable, Type, Iterator, Union, Tuple, Callable, Optional, Any, Dict

import sssom
from oaklib.datamodels.obograph import Node
from oaklib.datamodels.validation_datamodel import ValidationConfiguration, ValidationResult
from oaklib.interfaces.basic_ontology_interface import BasicOntologyInterface, RELATIONSHIP_MAP, PRED_CURIE, ALIAS_MAP, \
    METADATA_MAP, PREFIX_MAP
from oaklib.interfaces.mapping_provider_interface import MappingProviderInterface
from oaklib.interfaces.obograph_interface import OboGraphInterface
from oaklib.interfaces.search_interface import SearchInterface, SearchConfiguration
from oaklib.interfaces.validator_interface import ValidatorInterface
from oaklib.interfaces.rdf_interface import RdfInterface
from oaklib.interfaces.relation_graph_interface import RelationGraphInterface
from oaklib.types import CURIE, SUBSET_CURIE



@dataclass
class AggregatorImplementation(ValidatorInterface, RdfInterface, RelationGraphInterface, OboGraphInterface, SearchInterface, MappingProviderInterface):
    """
    """
    implementations: List[BasicOntologyInterface] = None

    def _delegate_iterator(self, func: Callable) -> Iterable:
        for i in self.implementations:
            for v in func(i):
                yield v

    def _delegate_simple_tuple_map(self, func: Callable, strict=False) -> Dict[Any, List[Any]]:
        m = defaultdict(list)
        for i in self.implementations:
            for k, vs in func(i).items():
                m[k] += vs
        return m


    def _delegate_first(self, func: Callable, strict=False) -> Optional[Any]:
        for i in self.implementations:
            v = func(i)
            if v is not None:
                return v
        if strict:
            raise ValueError(f'No value for {func}')

    def basic_search(self, search_term: str, config: SearchConfiguration = None) -> Iterable[CURIE]:
        return self._delegate_iterator(lambda i: i.basic_search(search_term, config=config))

    def validate(self, configuration: ValidationConfiguration = None) -> Iterable[ValidationResult]:
        return self._delegate_iterator(lambda i: i.validate())

    def all_entity_curies(self) -> Iterable[CURIE]:
        return self._delegate_iterator(lambda i: i.all_entity_curies())

    def get_simple_mappings_by_curie(self, curie: CURIE) -> Iterable[Tuple[PRED_CURIE, CURIE]]:
        return self._delegate_iterator(lambda i: i.get_simple_mappings_by_curie(curie))

    def get_sssom_mappings_by_curie(self, curie: CURIE) -> Iterable[sssom.Mapping]:
        return self._delegate_iterator(lambda i: i.get_sssom_mappings_by_curie(curie))

    def get_label_by_curie(self, curie: CURIE) -> str:
        return self._delegate_first(lambda i: i.get_label_by_curie(curie))

    def alias_map_by_curie(self, curie: CURIE) -> ALIAS_MAP:
        return self._delegate_simple_tuple_map(lambda i: i.alias_map_by_curie(curie))

    def all_subset_curies(self) -> Iterable[SUBSET_CURIE]:
        return self._delegate_iterator(lambda i: i.all_subset_curies())

    def curies_by_subset(self, subset: SUBSET_CURIE) -> Iterable[CURIE]:
        return self._delegate_iterator(lambda i: i.curies_by_subset(subset))

    def node(self, curie: CURIE, strict=False) -> Node:
        # TODO: this implementation is ad-hoc
        # return the first node that has a label populated
        node: Node = None
        for i in self.implementations:
            if isinstance(i, OboGraphInterface):
                node = i.node(curie)
                if node.label:
                    return node
        return node

    def get_outgoing_relationships_by_curie(self, curie: CURIE) -> RELATIONSHIP_MAP:
        return self._delegate_simple_tuple_map(lambda i: i.get_outgoing_relationships_by_curie(curie))

    def get_incoming_relationships_by_curie(self, curie: CURIE) -> RELATIONSHIP_MAP:
        return self._delegate_simple_tuple_map(lambda i: i.get_incoming_relationships_by_curie(curie))












from dataclasses import dataclass, field
from typing import List

from linkml_runtime.dumpers import yaml_dumper
from oaklib.io.streaming_writer import StreamingWriter
import sssom.sssom_datamodel as sssom_dm
from sssom.sssom_document import MappingSetDocument
from sssom.util import to_mapping_set_dataframe
from sssom.writers import write_table


@dataclass
class StreamingSssomWriter(StreamingWriter):
    """
    pseudo-streaming writer for SSSOM
    """
    mappings : List[sssom_dm.Mapping] = field(default_factory=lambda: [])

    def emit(self, obj: sssom_dm.Mapping):
        self.mappings.append(obj)

    def close(self):
        mset = sssom_dm.MappingSet(mapping_set_id='temp',
                                   mappings=self.mappings,
                                   license='UNSPECIFIED')
        doc = MappingSetDocument(prefix_map={}, mapping_set=mset)
        msdf = to_mapping_set_dataframe(doc)
        write_table(msdf, self.file)



# Auto generated from oxo.yaml by pythongen.py version: 0.9.0
# Generation date: 2022-04-12T11:15:58
# Schema: oxo-schema
#
# id: https://w3id.org/sssom/oxo
# description: Schema for OLS OXO payloads
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import sys
import re
from jsonasobj2 import JsonObj, as_dict
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
from linkml_runtime.linkml_model.meta import EnumDefinition, PermissibleValue, PvFormulaOptions

from linkml_runtime.utils.slot import Slot
from linkml_runtime.utils.metamodelcore import empty_list, empty_dict, bnode
from linkml_runtime.utils.yamlutils import YAMLRoot, extended_str, extended_float, extended_int
from linkml_runtime.utils.dataclass_extensions_376 import dataclasses_init_fn_with_kwargs
from linkml_runtime.utils.formatutils import camelcase, underscore, sfx
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from rdflib import Namespace, URIRef
from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_runtime.linkml_model.types import Integer, String, Uriorcurie
from linkml_runtime.utils.metamodelcore import URIorCURIE

metamodel_version = "1.7.0"
version = None

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
LINKML = CurieNamespace('linkml', 'https://w3id.org/linkml/')
OXO = CurieNamespace('oxo', 'https://w3id.org/sssom/oxo/')
XSD = CurieNamespace('xsd', 'http://www.w3.org/2001/XMLSchema#')
DEFAULT_ = OXO


# Types
class HttpsIdentifier(String):
    type_class_uri = XSD.string
    type_class_curie = "xsd:string"
    type_name = "https identifier"
    type_model_uri = OXO.HttpsIdentifier


class OntologyIdentifier(Uriorcurie):
    type_class_uri = XSD.anyURI
    type_class_curie = "xsd:anyURI"
    type_name = "OntologyIdentifier"
    type_model_uri = OXO.OntologyIdentifier


class HttpIdentifier(String):
    type_class_uri = XSD.string
    type_class_curie = "xsd:string"
    type_name = "http identifier"
    type_model_uri = OXO.HttpIdentifier


class Identifier(String):
    type_class_uri = XSD.string
    type_class_curie = "xsd:string"
    type_name = "identifier"
    type_model_uri = OXO.Identifier


# Class references
class TermCurie(extended_str):
    pass


@dataclass
class Datasource(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OXO.Datasource
    class_class_curie: ClassVar[str] = "oxo:Datasource"
    class_name: ClassVar[str] = "Datasource"
    class_model_uri: ClassVar[URIRef] = OXO.Datasource

    prefix: Optional[str] = None
    preferredPrefix: Optional[str] = None
    idorgNamespace: Optional[str] = None
    alternatePrefix: Optional[Union[str, List[str]]] = empty_list()
    alternateIris: Optional[Union[str, List[str]]] = empty_list()
    name: Optional[str] = None
    orcid: Optional[str] = None
    description: Optional[str] = None
    source: Optional[Union[str, "SourceEnum"]] = None
    licence: Optional[Union[str, HttpsIdentifier]] = None
    versionInfo: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.prefix is not None and not isinstance(self.prefix, str):
            self.prefix = str(self.prefix)

        if self.preferredPrefix is not None and not isinstance(self.preferredPrefix, str):
            self.preferredPrefix = str(self.preferredPrefix)

        if self.idorgNamespace is not None and not isinstance(self.idorgNamespace, str):
            self.idorgNamespace = str(self.idorgNamespace)

        if not isinstance(self.alternatePrefix, list):
            self.alternatePrefix = [self.alternatePrefix] if self.alternatePrefix is not None else []
        self.alternatePrefix = [v if isinstance(v, str) else str(v) for v in self.alternatePrefix]

        if not isinstance(self.alternateIris, list):
            self.alternateIris = [self.alternateIris] if self.alternateIris is not None else []
        self.alternateIris = [v if isinstance(v, str) else str(v) for v in self.alternateIris]

        if self.name is not None and not isinstance(self.name, str):
            self.name = str(self.name)

        if self.orcid is not None and not isinstance(self.orcid, str):
            self.orcid = str(self.orcid)

        if self.description is not None and not isinstance(self.description, str):
            self.description = str(self.description)

        if self.source is not None and not isinstance(self.source, SourceEnum):
            self.source = SourceEnum(self.source)

        if self.licence is not None and not isinstance(self.licence, HttpsIdentifier):
            self.licence = HttpsIdentifier(self.licence)

        if self.versionInfo is not None and not isinstance(self.versionInfo, str):
            self.versionInfo = str(self.versionInfo)

        super().__post_init__(**kwargs)


@dataclass
class Term(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OXO.Term
    class_class_curie: ClassVar[str] = "oxo:Term"
    class_name: ClassVar[str] = "Term"
    class_model_uri: ClassVar[URIRef] = OXO.Term

    curie: Union[str, TermCurie] = None
    identifier: Optional[str] = None
    uri: Optional[Union[str, HttpIdentifier]] = None
    label: Optional[str] = None
    datasource: Optional[Union[dict, Datasource]] = None
    href: Optional[Union[str, HttpsIdentifier]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.curie):
            self.MissingRequiredField("curie")
        if not isinstance(self.curie, TermCurie):
            self.curie = TermCurie(self.curie)

        if self.identifier is not None and not isinstance(self.identifier, str):
            self.identifier = str(self.identifier)

        if self.uri is not None and not isinstance(self.uri, HttpIdentifier):
            self.uri = HttpIdentifier(self.uri)

        if self.label is not None and not isinstance(self.label, str):
            self.label = str(self.label)

        if self.datasource is not None and not isinstance(self.datasource, Datasource):
            self.datasource = Datasource(**as_dict(self.datasource))

        if self.href is not None and not isinstance(self.href, HttpsIdentifier):
            self.href = HttpsIdentifier(self.href)

        super().__post_init__(**kwargs)


@dataclass
class Link(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OXO.Link
    class_class_curie: ClassVar[str] = "oxo:Link"
    class_name: ClassVar[str] = "Link"
    class_model_uri: ClassVar[URIRef] = OXO.Link

    href: Optional[Union[str, HttpsIdentifier]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.href is not None and not isinstance(self.href, HttpsIdentifier):
            self.href = HttpsIdentifier(self.href)

        super().__post_init__(**kwargs)


@dataclass
class LinkSet(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OXO.LinkSet
    class_class_curie: ClassVar[str] = "oxo:LinkSet"
    class_name: ClassVar[str] = "LinkSet"
    class_model_uri: ClassVar[URIRef] = OXO.LinkSet

    link_to_self: Optional[Union[dict, Link]] = None
    fromTerm: Optional[Union[dict, Link]] = None
    toTerm: Optional[Union[dict, Link]] = None
    first: Optional[Union[dict, Link]] = None
    next: Optional[Union[dict, Link]] = None
    last: Optional[Union[dict, Link]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.link_to_self is not None and not isinstance(self.link_to_self, Link):
            self.link_to_self = Link(**as_dict(self.link_to_self))

        if self.fromTerm is not None and not isinstance(self.fromTerm, Link):
            self.fromTerm = Link(**as_dict(self.fromTerm))

        if self.toTerm is not None and not isinstance(self.toTerm, Link):
            self.toTerm = Link(**as_dict(self.toTerm))

        if self.first is not None and not isinstance(self.first, Link):
            self.first = Link(**as_dict(self.first))

        if self.next is not None and not isinstance(self.next, Link):
            self.next = Link(**as_dict(self.next))

        if self.last is not None and not isinstance(self.last, Link):
            self.last = Link(**as_dict(self.last))

        super().__post_init__(**kwargs)


@dataclass
class Mapping(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OXO.Mapping
    class_class_curie: ClassVar[str] = "oxo:Mapping"
    class_name: ClassVar[str] = "Mapping"
    class_model_uri: ClassVar[URIRef] = OXO.Mapping

    mappingId: Optional[int] = None
    datasource: Optional[Union[dict, Datasource]] = None
    sourcePrefix: Optional[str] = None
    sourceType: Optional[Union[str, "SourceEnum"]] = None
    predicate: Optional[str] = None
    fromTerm: Optional[Union[dict, Term]] = None
    toTerm: Optional[Union[dict, Term]] = None
    scope: Optional[Union[str, "ScopeEnum"]] = None
    date: Optional[str] = None
    _links: Optional[Union[dict, LinkSet]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.mappingId is not None and not isinstance(self.mappingId, int):
            self.mappingId = int(self.mappingId)

        if self.datasource is not None and not isinstance(self.datasource, Datasource):
            self.datasource = Datasource(**as_dict(self.datasource))

        if self.sourcePrefix is not None and not isinstance(self.sourcePrefix, str):
            self.sourcePrefix = str(self.sourcePrefix)

        if self.sourceType is not None and not isinstance(self.sourceType, SourceEnum):
            self.sourceType = SourceEnum(self.sourceType)

        if self.predicate is not None and not isinstance(self.predicate, str):
            self.predicate = str(self.predicate)

        if self.fromTerm is not None and not isinstance(self.fromTerm, Term):
            self.fromTerm = Term(**as_dict(self.fromTerm))

        if self.toTerm is not None and not isinstance(self.toTerm, Term):
            self.toTerm = Term(**as_dict(self.toTerm))

        if self.scope is not None and not isinstance(self.scope, ScopeEnum):
            self.scope = ScopeEnum(self.scope)

        if self.date is not None and not isinstance(self.date, str):
            self.date = str(self.date)

        if self._links is not None and not isinstance(self._links, LinkSet):
            self._links = LinkSet(**as_dict(self._links))

        super().__post_init__(**kwargs)


@dataclass
class Embedded(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OXO.Embedded
    class_class_curie: ClassVar[str] = "oxo:Embedded"
    class_name: ClassVar[str] = "Embedded"
    class_model_uri: ClassVar[URIRef] = OXO.Embedded

    mappings: Optional[Union[Union[dict, Mapping], List[Union[dict, Mapping]]]] = empty_list()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if not isinstance(self.mappings, list):
            self.mappings = [self.mappings] if self.mappings is not None else []
        self.mappings = [v if isinstance(v, Mapping) else Mapping(**as_dict(v)) for v in self.mappings]

        super().__post_init__(**kwargs)


@dataclass
class Page(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OXO.Page
    class_class_curie: ClassVar[str] = "oxo:Page"
    class_name: ClassVar[str] = "Page"
    class_model_uri: ClassVar[URIRef] = OXO.Page

    size: Optional[int] = None
    totalElements: Optional[int] = None
    totalPages: Optional[int] = None
    number: Optional[int] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.size is not None and not isinstance(self.size, int):
            self.size = int(self.size)

        if self.totalElements is not None and not isinstance(self.totalElements, int):
            self.totalElements = int(self.totalElements)

        if self.totalPages is not None and not isinstance(self.totalPages, int):
            self.totalPages = int(self.totalPages)

        if self.number is not None and not isinstance(self.number, int):
            self.number = int(self.number)

        super().__post_init__(**kwargs)


@dataclass
class Container(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OXO.Container
    class_class_curie: ClassVar[str] = "oxo:Container"
    class_name: ClassVar[str] = "Container"
    class_model_uri: ClassVar[URIRef] = OXO.Container

    _embedded: Optional[Union[dict, Embedded]] = None
    _links: Optional[Union[dict, LinkSet]] = None
    page: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._embedded is not None and not isinstance(self._embedded, Embedded):
            self._embedded = Embedded(**as_dict(self._embedded))

        if self._links is not None and not isinstance(self._links, LinkSet):
            self._links = LinkSet(**as_dict(self._links))

        if self.page is not None and not isinstance(self.page, str):
            self.page = str(self.page)

        super().__post_init__(**kwargs)


# Enumerations
class SourceEnum(EnumDefinitionImpl):

    ONTOLOGY = PermissibleValue(text="ONTOLOGY")
    DATABASE = PermissibleValue(text="DATABASE")

    _defn = EnumDefinition(
        name="SourceEnum",
    )

class ScopeEnum(EnumDefinitionImpl):

    RELATED = PermissibleValue(text="RELATED")
    EXACT = PermissibleValue(text="EXACT")
    BROADER = PermissibleValue(text="BROADER")
    NARROWER = PermissibleValue(text="NARROWER")
    LABEL = PermissibleValue(text="LABEL")

    _defn = EnumDefinition(
        name="ScopeEnum",
    )

# Slots
class slots:
    pass

slots.prefix = Slot(uri=OXO.prefix, name="prefix", curie=OXO.curie('prefix'),
                   model_uri=OXO.prefix, domain=None, range=Optional[str])

slots.preferredPrefix = Slot(uri=OXO.preferredPrefix, name="preferredPrefix", curie=OXO.curie('preferredPrefix'),
                   model_uri=OXO.preferredPrefix, domain=None, range=Optional[str])

slots.idorgNamespace = Slot(uri=OXO.idorgNamespace, name="idorgNamespace", curie=OXO.curie('idorgNamespace'),
                   model_uri=OXO.idorgNamespace, domain=None, range=Optional[str])

slots.alternatePrefix = Slot(uri=OXO.alternatePrefix, name="alternatePrefix", curie=OXO.curie('alternatePrefix'),
                   model_uri=OXO.alternatePrefix, domain=None, range=Optional[Union[str, List[str]]])

slots.alternateIris = Slot(uri=OXO.alternateIris, name="alternateIris", curie=OXO.curie('alternateIris'),
                   model_uri=OXO.alternateIris, domain=None, range=Optional[Union[str, List[str]]])

slots.name = Slot(uri=OXO.name, name="name", curie=OXO.curie('name'),
                   model_uri=OXO.name, domain=None, range=Optional[str])

slots.orcid = Slot(uri=OXO.orcid, name="orcid", curie=OXO.curie('orcid'),
                   model_uri=OXO.orcid, domain=None, range=Optional[str])

slots.description = Slot(uri=OXO.description, name="description", curie=OXO.curie('description'),
                   model_uri=OXO.description, domain=None, range=Optional[str])

slots.source = Slot(uri=OXO.source, name="source", curie=OXO.curie('source'),
                   model_uri=OXO.source, domain=None, range=Optional[Union[str, "SourceEnum"]])

slots.licence = Slot(uri=OXO.licence, name="licence", curie=OXO.curie('licence'),
                   model_uri=OXO.licence, domain=None, range=Optional[Union[str, HttpsIdentifier]])

slots.versionInfo = Slot(uri=OXO.versionInfo, name="versionInfo", curie=OXO.curie('versionInfo'),
                   model_uri=OXO.versionInfo, domain=None, range=Optional[str])

slots.curie = Slot(uri=OXO.curie, name="curie", curie=OXO.curie('curie'),
                   model_uri=OXO.curie, domain=None, range=URIRef)

slots.identifier = Slot(uri=OXO.identifier, name="identifier", curie=OXO.curie('identifier'),
                   model_uri=OXO.identifier, domain=None, range=Optional[str])

slots.uri = Slot(uri=OXO.uri, name="uri", curie=OXO.curie('uri'),
                   model_uri=OXO.uri, domain=None, range=Optional[Union[str, HttpIdentifier]])

slots.label = Slot(uri=OXO.label, name="label", curie=OXO.curie('label'),
                   model_uri=OXO.label, domain=None, range=Optional[str])

slots.datasource = Slot(uri=OXO.datasource, name="datasource", curie=OXO.curie('datasource'),
                   model_uri=OXO.datasource, domain=None, range=Optional[Union[dict, Datasource]])

slots.href = Slot(uri=OXO.href, name="href", curie=OXO.curie('href'),
                   model_uri=OXO.href, domain=None, range=Optional[Union[str, HttpsIdentifier]])

slots.first = Slot(uri=OXO.first, name="first", curie=OXO.curie('first'),
                   model_uri=OXO.first, domain=None, range=Optional[Union[dict, Link]])

slots.next = Slot(uri=OXO.next, name="next", curie=OXO.curie('next'),
                   model_uri=OXO.next, domain=None, range=Optional[Union[dict, Link]])

slots.last = Slot(uri=OXO.last, name="last", curie=OXO.curie('last'),
                   model_uri=OXO.last, domain=None, range=Optional[Union[dict, Link]])

slots.link_to_self = Slot(uri=OXO.link_to_self, name="link_to_self", curie=OXO.curie('link_to_self'),
                   model_uri=OXO.link_to_self, domain=None, range=Optional[Union[dict, Link]])

slots.fromTerm = Slot(uri=OXO.fromTerm, name="fromTerm", curie=OXO.curie('fromTerm'),
                   model_uri=OXO.fromTerm, domain=None, range=Optional[Union[str, TermCurie]])

slots.toTerm = Slot(uri=OXO.toTerm, name="toTerm", curie=OXO.curie('toTerm'),
                   model_uri=OXO.toTerm, domain=None, range=Optional[Union[str, TermCurie]])

slots.mappingId = Slot(uri=OXO.mappingId, name="mappingId", curie=OXO.curie('mappingId'),
                   model_uri=OXO.mappingId, domain=None, range=Optional[int])

slots.sourcePrefix = Slot(uri=OXO.sourcePrefix, name="sourcePrefix", curie=OXO.curie('sourcePrefix'),
                   model_uri=OXO.sourcePrefix, domain=None, range=Optional[str])

slots.sourceType = Slot(uri=OXO.sourceType, name="sourceType", curie=OXO.curie('sourceType'),
                   model_uri=OXO.sourceType, domain=None, range=Optional[Union[str, "SourceEnum"]])

slots.predicate = Slot(uri=OXO.predicate, name="predicate", curie=OXO.curie('predicate'),
                   model_uri=OXO.predicate, domain=None, range=Optional[str])

slots.scope = Slot(uri=OXO.scope, name="scope", curie=OXO.curie('scope'),
                   model_uri=OXO.scope, domain=None, range=Optional[Union[str, "ScopeEnum"]])

slots.date = Slot(uri=OXO.date, name="date", curie=OXO.curie('date'),
                   model_uri=OXO.date, domain=None, range=Optional[str])

slots._links = Slot(uri=OXO._links, name="_links", curie=OXO.curie('_links'),
                   model_uri=OXO._links, domain=None, range=Optional[Union[dict, LinkSet]])

slots.mappings = Slot(uri=OXO.mappings, name="mappings", curie=OXO.curie('mappings'),
                   model_uri=OXO.mappings, domain=None, range=Optional[Union[Union[dict, Mapping], List[Union[dict, Mapping]]]])

slots.size = Slot(uri=OXO.size, name="size", curie=OXO.curie('size'),
                   model_uri=OXO.size, domain=None, range=Optional[int])

slots.totalElements = Slot(uri=OXO.totalElements, name="totalElements", curie=OXO.curie('totalElements'),
                   model_uri=OXO.totalElements, domain=None, range=Optional[int])

slots.totalPages = Slot(uri=OXO.totalPages, name="totalPages", curie=OXO.curie('totalPages'),
                   model_uri=OXO.totalPages, domain=None, range=Optional[int])

slots.number = Slot(uri=OXO.number, name="number", curie=OXO.curie('number'),
                   model_uri=OXO.number, domain=None, range=Optional[int])

slots._embedded = Slot(uri=OXO._embedded, name="_embedded", curie=OXO.curie('_embedded'),
                   model_uri=OXO._embedded, domain=None, range=Optional[Union[dict, Embedded]])

slots.page = Slot(uri=OXO.page, name="page", curie=OXO.curie('page'),
                   model_uri=OXO.page, domain=None, range=Optional[str])

slots.LinkSet_fromTerm = Slot(uri=OXO.fromTerm, name="LinkSet_fromTerm", curie=OXO.curie('fromTerm'),
                   model_uri=OXO.LinkSet_fromTerm, domain=LinkSet, range=Optional[Union[dict, Link]])

slots.LinkSet_toTerm = Slot(uri=OXO.toTerm, name="LinkSet_toTerm", curie=OXO.curie('toTerm'),
                   model_uri=OXO.LinkSet_toTerm, domain=LinkSet, range=Optional[Union[dict, Link]])

slots.LinkSet_first = Slot(uri=OXO.first, name="LinkSet_first", curie=OXO.curie('first'),
                   model_uri=OXO.LinkSet_first, domain=LinkSet, range=Optional[Union[dict, Link]])

slots.LinkSet_next = Slot(uri=OXO.next, name="LinkSet_next", curie=OXO.curie('next'),
                   model_uri=OXO.LinkSet_next, domain=LinkSet, range=Optional[Union[dict, Link]])

slots.LinkSet_last = Slot(uri=OXO.last, name="LinkSet_last", curie=OXO.curie('last'),
                   model_uri=OXO.LinkSet_last, domain=LinkSet, range=Optional[Union[dict, Link]])

slots.Mapping_fromTerm = Slot(uri=OXO.fromTerm, name="Mapping_fromTerm", curie=OXO.curie('fromTerm'),
                   model_uri=OXO.Mapping_fromTerm, domain=Mapping, range=Optional[Union[dict, Term]])

slots.Mapping_toTerm = Slot(uri=OXO.toTerm, name="Mapping_toTerm", curie=OXO.curie('toTerm'),
                   model_uri=OXO.Mapping_toTerm, domain=Mapping, range=Optional[Union[dict, Term]])



from abc import ABC
from typing import Dict, List, Tuple, Iterable, Union

from oaklib.interfaces.basic_ontology_interface import BasicOntologyInterface, RELATIONSHIP_MAP, RELATIONSHIP
from oaklib.types import CURIE, LABEL, URI, PRED_CURIE
from oaklib.utilities.graph.relationship_walker import walk_up


class RelationGraphInterface(BasicOntologyInterface, ABC):
    """
    an interface that provides relation graph abstractions

    .. note ::

        that the operations provided here are similar to the graph-walking operations provided in :class:`.OboGraphInterface`.
        The main difference is that a RG provides a more restricted and formally correct set of entailments
    """

    def entailed_outgoing_relationships_by_curie(self, curie: CURIE,
                                                 predicates: List[PRED_CURIE] = None) -> Iterable[Tuple[PRED_CURIE, CURIE]]:
        """
        The return relationship map is keyed by relationship type, where the values
        are the 'parents' or fillers

        OWL formulation:

         - is_a: {P : SubClassOf(C P), class(P)}
         - R: {P : SubClassOf(C ObjectSomeValuesFrom( RP), class(P), property(P)}

        :param curie: the 'child' term
        :param predicates:
        :return:
        """
        raise NotImplementedError

    def entailed_incoming_relationships_by_curie(self, curie: CURIE,
                                                 predicates: List[PRED_CURIE] = None) -> Iterable[Tuple[PRED_CURIE, CURIE]]:
        """
        The return relationship map is keyed by relationship type, where the values
        are the 'parents' or fillers

        OWL formulation:

         - is_a: {P : SubClassOf(C P), class(P)}
         - R: {P : SubClassOf(C ObjectSomeValuesFrom( RP), class(P), property(P)}

        :param curie: the 'child' term
        :param predicates:
        :return:
        """
        raise NotImplementedError

    def walk_up_relationship_graph(self, start_curies: Union[CURIE, List[CURIE]], predicates: List[PRED_CURIE] = None) -> Iterable[RELATIONSHIP]:
        """
        Walks up the relation graph from a seed set of curies or individual curie, returning the full ancestry graph

        Note: this may be inefficient for remote endpoints, in future a graph walking endpoint will implement this

        :param start_curies:
        :param predicates:
        :return:
        """
        return walk_up(self, start_curies, predicates=predicates)

    def xxxancestors(self, start_curies: Union[CURIE, List[CURIE]], predicates: List[PRED_CURIE] = None) -> Iterable[CURIE]:
        # TODO: make reflexivity a parameters
        ancs = set([x[2] for x in self.walk_up_relationship_graph(start_curies, predicates)] + [start_curies])
        for a in ancs:
            yield a





from abc import ABC
from dataclasses import dataclass
from typing import Dict, List, Tuple, Iterator

from oaklib.interfaces.basic_ontology_interface import BasicOntologyInterface, RELATIONSHIP_MAP, RELATIONSHIP
from oaklib.types import CURIE, LABEL, URI, PRED_CURIE


@dataclass
class SubsetStrategy:
    pass


class SyntacticLocalityModuleExtraction(SubsetStrategy):
    module_type: str ## TODO enum


class GraphWalkExtraction(SubsetStrategy):
    predicates: List[PRED_CURIE]


class MireotExtraction(SubsetStrategy):
    pass


class SubsetterInterface(BasicOntologyInterface, ABC):
    """
    an interface that provides subsetting operations

    a challenge here is what a subset operation should produce:

     - an ontology object (committing to a specific object model)
     - an ontology interface
     - a reference to an ontology
    """

    def extract_subset_ontology(self, seed_curies: List[CURIE], strategy: SubsetStrategy = None) -> BasicOntologyInterface:
        """

        :param seed_curies:
        :param strategy:
        :return:
        """
        raise NotImplementedError

    def gap_fill_relationships(self, seed_curies: List[CURIE], predicates: List[PRED_CURIE] = None) -> Iterator[RELATIONSHIP]:
        """
        Given a term subset as a list of curies, find all non-redundant relationships connecting them

        This assumes relation-graph entailed edges, so currently only implemented for ubergraph and sqlite

        First the subset of all entailed edges conforming to the predicate profile connecting any pair of terms in the
        subset is selected

        Then naive transitive reduction on a per predicate basis is performed. This may yield edges that are formally
        redundant, but these are still assumed to be useful for the user

        :param seed_curies:
        :param predicates: if specified, only consider relationships using these predicates
        :return:
        """
        raise NotImplementedError






import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Type, Union

from oaklib.implementations.bioportal.bioportal_implementation import BioportalImplementation
from oaklib.implementations.ols.ols_implementation import OlsImplementation
from oaklib.implementations.ontobee.ontobee_implementation import OntobeeImplementation
from oaklib.implementations.pronto.pronto_implementation import ProntoImplementation
from oaklib.implementations.sqldb.sql_implementation import SqlImplementation
from oaklib.implementations.ubergraph import UbergraphImplementation
from oaklib.interfaces import OntologyInterface
from oaklib.resource import OntologyResource


def get_implementation_from_shorthand(descriptor: str, format: str = None) -> OntologyResource:
    """
    See :ref:`get_resource_from_shorthand`

    Example:

    .. code :: python

        >>> oi = get_implementation_from_shorthand('my-ont.owl')
        >>> for term in oi.all_entities():
        >>>     ...

    :param descriptor:
    :param format:
    :return:
    """
    res = get_resource_from_shorthand(descriptor, format)
    return res.implementation_class(res)

def get_resource_from_shorthand(descriptor: str, format: str = None) -> OntologyResource:
    """
    Maps from a shorthand descriptor to an OntologyResource.




    :param descriptor:
    :param format:
    :return:
    """
    resource = OntologyResource(format=format)
    resource.slug = descriptor
    impl_class: Type[OntologyInterface]
    if descriptor:
        if ':' in descriptor:
            toks = descriptor.split(':')
            scheme = toks[0]
            rest = ':'.join(toks[1:])
            if not rest:
                rest= None
            resource.slug = rest
            if scheme == 'sqlite':
                impl_class = SqlImplementation
                resource.slug = f'sqlite:///{Path(rest).absolute()}'
            elif scheme == 'ubergraph':
                impl_class = UbergraphImplementation
            elif scheme == 'ontobee':
                impl_class = OntobeeImplementation
            elif scheme == 'bioportal':
                impl_class = BioportalImplementation
            elif scheme == 'ols':
                impl_class = OlsImplementation
            elif scheme == 'pronto':
                impl_class = ProntoImplementation
                if resource.slug.endswith('.obo'):
                    resource.format = 'obo'
                resource.local = True
                resource.slug = rest
            elif scheme == 'obolibrary' or scheme == 'prontolib':
                impl_class = ProntoImplementation
                if resource.slug.endswith('.obo'):
                    resource.format = 'obo'
                resource.local = False
                resource.slug = rest
            elif scheme == 'http' or scheme == 'https':
                raise NotImplementedError(f'Web requests not implemented yet')
            else:
                raise ValueError(f'Scheme {scheme} not known')
        else:
            logging.info(f'No schema: assuming file path {descriptor}')
            if descriptor.endswith('.db') or (format and format == 'sqlite'):
                impl_class = SqlImplementation
                resource.slug = f'sqlite:///{Path(descriptor).absolute()}'
            else:
                resource.local = True
                impl_class = ProntoImplementation
    else:
        raise ValueError(f'No descriptor')
    resource.implementation_class = impl_class
    return resource



"""
https://stackoverflow.com/questions/40748687/python-api-rate-limiting-how-to-limit-api-calls-globally
"""
from ratelimit import limits, sleep_and_retry

# calls per minute
CALLS = 300
RATE_LIMIT = 60
#CALLS = 1
#RATE_LIMIT = 10

@sleep_and_retry
@limits(calls=CALLS, period=RATE_LIMIT)
def check_limit():
    """ Empty function just to check for calls to API """
    return


from abc import ABC
from typing import Dict, List, Iterable

from oaklib.interfaces.basic_ontology_interface import BasicOntologyInterface
from oaklib.types import CURIE, LABEL, URI, PRED_CURIE


class SemanticSimilarityInterface(BasicOntologyInterface, ABC):
    """
    TODO: consider direct use of nxontology
    """

    def most_recent_common_ancestors(self, curie1: CURIE, curie2: CURIE, predicates: List[PRED_CURIE] = None) -> Iterable[CURIE]:
        raise NotImplementedError

    def common_ancestors(self, curie1: CURIE, curie2: CURIE, predicates: List[PRED_CURIE] = None) -> Iterable[CURIE]:
        raise NotImplementedError

    def get_information_content(self, curie: CURIE, background: CURIE = None):
        raise NotImplementedError




import logging
from collections import defaultdict
from dataclasses import dataclass
from functools import lru_cache
from typing import List, Iterable, Tuple, Optional, Union, Iterator

import SPARQLWrapper
import rdflib
import sssom
from SPARQLWrapper import JSON
from oaklib.implementations.sparql.sparql_query import SparqlQuery
from oaklib.interfaces.basic_ontology_interface import BasicOntologyInterface, RELATIONSHIP_MAP, PRED_CURIE, ALIAS_MAP, \
    PREFIX_MAP
from oaklib.interfaces.mapping_provider_interface import MappingProviderInterface
from oaklib.interfaces.rdf_interface import TRIPLE, RdfInterface
from oaklib.interfaces.search_interface import SearchConfiguration
from oaklib.resource import OntologyResource
from oaklib.types import CURIE, URI
from oaklib.datamodels.vocabulary import IS_A, HAS_DEFINITION_URI, LABEL_PREDICATE, OBO_PURL, ALL_MATCH_PREDICATES, \
    DEFAULT_PREFIX_MAP
from oaklib.utilities.rate_limiter import check_limit
from rdflib import URIRef, RDFS, Literal, BNode
from sssom.sssom_datamodel import MatchTypeEnum

VAL_VAR = 'v'

def _sparql_values(var_name: str, vals: List[str]):
    return f'VALUES ?{var_name} {{ {" ".join(vals)} }}'

def _as_rdf_obj(v) -> URIRef:
    val = v['value']
    if v['type'] == 'uri':
        return URIRef(val)
    elif v['type'] == 'bnode':
        return BNode(val)
    else:
        return Literal(val, lang=v.get('lang', None), datatype=v.get('datatype', None))

@dataclass
class SparqlImplementation(RdfInterface):
    """
    An OntologyInterface implementation that wraps a (typically remote) SPARQL endpoint.

    Note: Each sparql endpoint typically implements its own bespoke transformations, leading
    to lack of interoperability. For some purposes it may be better to use a more specific
    implementation:

    - :class:`.OntobeeImplementation`
    - :class:`.UbergraphImplementation`
    """
    sparql_wrapper: SPARQLWrapper = None
    graph: rdflib.Graph = None
    _list_of_named_graphs: List[str] = None

    def __post_init__(self):
        if self.sparql_wrapper is None:
            resource = self.resource
            if resource is None:
                resource = OntologyResource()
            if resource.url is None:
                resource.url = self._default_url()
            if resource.local:
                self.graph = rdflib.Graph()
                self.graph.parse(resource.local_path)
            else:
                self.sparql_wrapper = SPARQLWrapper.SPARQLWrapper(resource.url)

    @property
    def named_graph(self) -> Optional[str]:
        return None

    def _default_url(self) -> str:
        raise NotImplementedError

    def _is_blazegraph(self) -> bool:
        return False



    def get_prefix_map(self) -> PREFIX_MAP:
        # TODO
        #return {'rdfs': str(RDFS)}
        return DEFAULT_PREFIX_MAP

    #def store(self, resource: OntologyResource) -> None:
    #    SparqlBasicImpl.dump(self.engine, resource)

    def curie_to_uri(self, curie: CURIE, strict: bool = False) -> URI:
        if curie.startswith('http'):
            return curie
        pm = self.get_prefix_map()
        if ':' in curie:
            pfx, local_id = curie.split(':')
            if pfx in pm:
                return f'{pm[pfx]}{local_id}'
            else:
                return f'http://purl.obolibrary.org/obo/{pfx}_{local_id}'
        else:
            logging.warning(f'Not a curie: {curie}')
            return curie

    def curie_to_sparql(self, curie: CURIE, strict: bool = False) -> URI:
        return f'<{self.curie_to_uri(curie, strict=strict)}>'

    def uri_to_curie(self, uri: URI, strict=True) -> Optional[CURIE]:
        # TODO: do not hardcode OBO
        pm = self.get_prefix_map()
        for k, v in pm.items():
            if uri.startswith(v):
                return uri.replace(v, f'{k}:')
        if uri.startswith(OBO_PURL):
            uri = uri.replace(OBO_PURL, "")
            return uri.replace('_', ':')
        return uri

    def list_of_named_graphs(self) -> List[URI]:
        if self._list_of_named_graphs:
            return self._list_of_named_graphs
        query = "select DISTINCT ?g WHERE { GRAPH ?g { ?s ?p ?o }}"
        sw = self.sparql_wrapper
        sw.setQuery(query)
        sw.setReturnFormat(JSON)
        check_limit()
        ret = sw.queryAndConvert()
        logging.info(f'RET={ret}')
        self._list_of_named_graphs = [row['g']['value'] for row in ret["results"]["bindings"]]
        return self._list_of_named_graphs

    def _query(self, query: Union[str, SparqlQuery], prefixes: PREFIX_MAP = {}):
        ng = self.named_graph
        if isinstance(query, SparqlQuery) and ng:
            query.graph = ng
        if isinstance(query, SparqlQuery):
            query = query.query_str()
        sw = self.sparql_wrapper
        for k, v in prefixes.items():
            query = f'PREFIX {k}: <{v}>\n' + query
        logging.info(f'QUERY={query} // sw={sw}')
        sw.setQuery(query)
        sw.setReturnFormat(JSON)
        check_limit()
        ret = sw.queryAndConvert()
        logging.info(f'RET={ret}')
        return ret["results"]["bindings"]

    def _triples(self, subject: CURIE = None, predicate: PRED_CURIE = None, object: PRED_CURIE = None, graph: CURIE = None) -> Iterable[Tuple]:
        vars = []
        def _urify_arg(arg: CURIE, name: str):
            if arg is None:
                vars.append(name)
                return f'?{name}'
            else:
                return f'<{self.curie_to_uri(arg)}>'
        subject_uri = _urify_arg(subject, 's')
        predicate_uri = _urify_arg(predicate, 'p')
        object_uri = _urify_arg(object, 'o')
        if graph is None:
            # do not include graph in select
            graph_uri = '?g'
        else:
            graph_uri = _urify_arg(graph, 'g')
        vars_str = ' '.join([f'?{v}' for v in vars])
        bindings = self._query(f"SELECT DISTINCT {vars_str} WHERE {{ GRAPH {graph_uri} {{ {subject_uri} {predicate_uri} {object_uri} }}}}")
        for row in bindings:
            yield tuple([row[v]['value'] for v in vars])

    def get_parents_by_curie(self, curie: CURIE, isa_only: bool = False) -> List[CURIE]:
        uri = self.curie_to_uri(curie)
        query = SparqlQuery(select=['?o'],
                            where=[f'<{uri}> <{RDFS.subClassOf}> ?o',
                                   'FILTER (isIRI(?o))'])
        bindings = self._query(query)
        return list(set([self.uri_to_curie(row['o']['value']) for row in bindings]))

    def get_outgoing_relationships_by_curie(self, curie: CURIE, isa_only: bool = False) -> RELATIONSHIP_MAP:
        uri = self.curie_to_uri(curie)
        rels = defaultdict(list)
        rels[IS_A] = self.get_parents_by_curie(curie)
        query = SparqlQuery(select=['?p', '?o'],
                            where=[f'<{uri}> <{RDFS.subClassOf}> [owl:onProperty ?p ; owl:someValuesFrom ?o]',
                                   'FILTER (isIRI(?o))'])
        bindings = self._query(query)
        for row in bindings:
            pred = self.uri_to_curie(row['p']['value'])
            obj = self.uri_to_curie(row['o']['value'])
            if obj not in rels[pred]:
                rels[pred].append(obj)
        return rels


    def _get_anns(self, curie: CURIE, pred: URIRef):
        uri = self.curie_to_sparql(curie)
        query = SparqlQuery(select=['?v'],
                            where=[f'{uri} <{pred}> ?v'])
        #bindings = self._query(f"SELECT ?{VAL_VAR} WHERE {{ <{uri}> <{pred}> ?v }}")
        bindings = self._query(query)
        return list(set([row[VAL_VAR]['value'] for row in bindings]))

    def get_label_by_curie(self, curie: CURIE):
        labels = self._get_anns(curie, RDFS.label)
        if labels:
            if len(labels) > 1:
                logging.warning(f'Multiple labels for {curie} = {labels}')
            return labels[0]
        else:
            return None

    def get_labels_for_curies(self, curies: Iterable[CURIE]) -> Iterable[Tuple[CURIE, str]]:
        uri_map = {self.curie_to_uri(curie): curie for curie in curies}
        uris = [f'<{uri}>' for uri in uri_map.keys()]
        query = SparqlQuery(select=['?s ?label'],
                            where=[f'?s <{RDFS.label}> ?label',
                                   f'VALUES ?s {{ {" ".join(uris)} }}'])
        bindings = self._query(query)
        label_map = {}
        for row in bindings:
            curie, label = self.uri_to_curie(row['s']['value']), row['label']['value']
            if curie in label_map:
                if label_map[curie] != label:
                    logging.warning(f'Multiple labels for {curie} = {label_map[curie]} != {label}')
            else:
                yield curie, label
        for curie in curies:
            if curie not in label_map:
                yield curie, None

    def alias_map_by_curie(self, curie: CURIE) -> ALIAS_MAP:
        uri = self.curie_to_uri(curie)
        valstr = "VALUES ?pred {oboInOwl:hasExactSynonym}"
        bindings = self._query(f"SELECT ?pred ?{VAL_VAR} WHERE {{ <{uri}> ?pred ?v . {valstr} }}",
                               {'oboInOwl': 'http://www.geneontology.org/formats/oboInOwl#'})
        m = defaultdict(list)
        for row in bindings:
            #print(f'BINDINGS={row}')
            m[row['pred']['value']].append(row[VAL_VAR]['value'])
        return m


    def get_curies_by_label(self, label: str) -> List[CURIE]:
        query = SparqlQuery(select=['?s'],
                            where=[f'{{ {{ ?s rdfs:label "{label}" }} UNION {{ ?s rdfs:label "{label}"^^xsd:string }}  }}'])
        bindings = self._query(query)
        return [self.uri_to_curie(row['s']['value']) for row in bindings]


    def create_entity(self, curie: CURIE, label: str = None, relationships: RELATIONSHIP_MAP = None) -> CURIE:
        raise NotImplementedError

    def add_relationship(self, curie: CURIE, predicate: PRED_CURIE, filler: CURIE):
        raise NotImplementedError

    def get_definition_by_curie(self, curie: CURIE) -> str:
        """

        :param curie:
        :return:
        """
        labels = self._get_anns(curie, HAS_DEFINITION_URI)
        if labels:
            if len(labels) > 1:
                logging.error(f'Multiple labels for {curie} = {labels}')
            return labels[0]
        else:
            return None

    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    # Implements: SearchInterface
    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    def basic_search(self, search_term: str, config: SearchConfiguration = SearchConfiguration()) -> Iterable[CURIE]:
        if ':' in search_term and ' ' not in search_term:
            logging.debug(f'Not performing search on what looks like a CURIE: {search_term}')
            return
        if self._is_blazegraph():
            filter_clause = f'?v bds:search "{search_term}"'
        else:
            if config.complete:
                filter_clause = f'?v = "{search_term}"'
            else:
                filter_clause = f'strStarts(str(?v), "{search_term}")'
            filter_clause = f'FILTER({filter_clause})'
        preds = [LABEL_PREDICATE]
        if len(preds) == 1:
            where = [f'?s {preds[0]} ?v ']
        else:
            where = [f'?s ?p ?v ',
                     f'VALUES ?p {{ {" ".join(preds)} }}']
        query = SparqlQuery(select=['?s'],
                            where=where + [filter_clause])
        bindings = self._query(query)
        for row in bindings:
            yield self.uri_to_curie(row['s']['value'])


    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    # Implements: MappingProviderInterface
    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    def get_sssom_mappings_by_curie(self, curie: CURIE) -> Iterable[sssom.Mapping]:
        pred_uris = [self.curie_to_sparql(pred) for pred in ALL_MATCH_PREDICATES]
        query = SparqlQuery(select=['?p', '?o'],
                            where=[f'{self.curie_to_sparql(curie)} ?p ?o',
                                   f'VALUES ?p {{ {" ".join(pred_uris)} }}'
                                   ])
        bindings = self._query(query)
        for row in bindings:
            yield sssom.Mapping(subject_id=curie,
                                predicate_id=self.uri_to_curie(row['p']['value']),
                                object_id=self.uri_to_curie(row['o']['value']),
                                match_type=MatchTypeEnum.Unspecified,
                                )

    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    # Implements: RdfInterface
    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    def extract_triples(self, seed_curies: List[CURIE], predicates: List[PRED_CURIE] = None, strategy=None,
                        map_to_curies=True) -> Iterator[TRIPLE]:
        seed_uris = [self.curie_to_sparql(c) for c in seed_curies]
        query = SparqlQuery(select=['?s', '?p', '?o'],
                            where=['?s ?p ?o .'
                                   '?seed (rdfs:subClassOf|owl:onProperty|owl:someValuesFrom|^owl:annotatedSource)* ?s',
                                    _sparql_values('seed', seed_uris)])
        bindings = self._query(query)
        for row in bindings:
            triple = (row['s'], row['p'], row['o'])
            if map_to_curies:
                yield tuple([self.uri_to_curie(v['value']) for v in list(triple)])
            else:
                yield tuple([_as_rdf_obj(v) for v in list(triple)])










from abc import ABC
from typing import Iterator

from oaklib.datamodels.text_annotator import TextAnnotation
from oaklib.interfaces.basic_ontology_interface import BasicOntologyInterface

TEXT = str

class TextAnnotatorInterface(BasicOntologyInterface, ABC):
    """
    Performs Named Entity Recognition on texts

    Currently this is only partially implemented by :class:`.BioportalInterface`

    potential implementations:

    - zooma
    - scigraph-annotator
    - ontorunner
    - spacy
    """

    def annotate_text(self, text: TEXT) -> Iterator[TextAnnotation]:
        """
        Annotate a piece of text

        .. note ::

           the signature of this method may change

        :param text:
        :return:
        """
        raise NotImplementedError



from typing import List, Iterator

from oaklib.datamodels.vocabulary import IS_A
from oaklib.interfaces.mapping_provider_interface import MappingProviderInterface
from oaklib.interfaces.obograph_interface import OboGraphInterface
from oaklib.types import CURIE, PRED_CURIE


def propagate_terms_over_mappings(oi: MappingProviderInterface, curies: List[CURIE],
                                  predicates: List[PRED_CURIE] = [IS_A]) -> Iterator:
    if isinstance(oi, OboGraphInterface):
        for curie in curies:
            for anc in oi.ancestors(curie, predicates):
                if isinstance(oi, MappingProviderInterface):
                    for m in oi.get_sssom_mappings_by_curie(curie):
                        yield curie, anc, m.object_id


from typing import Any
import oaklib.datamodels.oxo as oxo
from linkml_runtime.loaders import json_loader


def fix_json_payload(obj: Any) -> None:
    """
    changes payload in place

    - self --> link_to_self

    :param obj:
    :return:
    """
    if isinstance(obj, dict):
        for k, v in obj.items():
            fix_json_payload(v)
        if 'self' in obj:
            obj['link_to_self'] = obj['self']
            del obj['self']
    elif isinstance(obj, list):
        [fix_json_payload(v) for v in obj]
    else:
        pass

def load_oxo_payload(obj: Any) -> oxo.Container:
    fix_json_payload(obj)
    return json_loader.loads(obj, target_class=oxo.Container)



"""
Lexical Utilities
-----------------

Various utilities for working with lexical aspects of ontologies plus mappings

"""
import logging
import re
from collections import defaultdict
from typing import List, Dict, Optional, Tuple

from linkml_runtime.dumpers import yaml_dumper
from linkml_runtime.loaders import yaml_loader
from oaklib.interfaces import BasicOntologyInterface
from oaklib.types import PRED_CURIE
from oaklib.datamodels.lexical_index import LexicalIndex, LexicalTransformation, TransformationType, RelationshipToTerm, \
    LexicalGrouping, LexicalTransformationPipeline
from oaklib.datamodels.mapping_rules_datamodel import Precondition, MappingRuleCollection
from oaklib.datamodels.vocabulary import SKOS_EXACT_MATCH, SKOS_BROAD_MATCH, SKOS_NARROW_MATCH, \
    SKOS_CLOSE_MATCH
from oaklib.utilities.basic_utils import pairs_as_dict
from sssom import Mapping
from sssom.sssom_document import MappingSetDocument
from sssom.util import MappingSetDataFrame, to_mapping_set_dataframe
from sssom.sssom_datamodel import MatchTypeEnum, MappingSet

def add_labels_from_uris(oi: BasicOntologyInterface):
    """
    Adds a label based on the CURIE or URI for entities that lack labels

    :param oi:
    :return:
    """
    for curie in oi.all_entity_curies():
        if not oi.get_label_by_curie(curie):
            if '#' in curie:
                sep = '#'
            elif curie.startswith('http'):
                sep = '/'
            else:
                sep = ':'
            label = curie.split(sep)[-1]
            oi.set_label_for_curie(curie, label)


def create_lexical_index(oi: BasicOntologyInterface,
                         pipelines: List[LexicalTransformationPipeline] = None) -> LexicalIndex:
    """
    Generates a LexicalIndex keyed by normalized terms

    If the pipelines parameter is not specified, then default pipelines will be applied
    (currently CaseNormalization and WhitespaceNormalization)

    :param oi:
    :param pipelines: list of transformation pipelines to apply
    :return:
    """
    if pipelines is None:
        step1 = LexicalTransformation(TransformationType.CaseNormalization)
        step2 = LexicalTransformation(TransformationType.WhitespaceNormalization)
        pipelines = [LexicalTransformationPipeline(name='default',
                                                   transformations=[step1, step2])]
    ix = LexicalIndex(pipelines={p.name: p for p in pipelines})
    for curie in oi.all_entity_curies():
        alias_map = oi.alias_map_by_curie(curie)
        mapping_map = pairs_as_dict(oi.get_simple_mappings_by_curie(curie))
        for pred, terms in {**alias_map, **mapping_map}.items():
            for term in terms:
                if not term:
                    logging.warning(f'No term for {curie}.{pred}')
                    continue
                for pipeline in pipelines:
                    term2 = term
                    for tr in pipeline.transformations:
                        term2 = apply_transformation(term, tr)
                    rel = RelationshipToTerm(predicate=pred, element=curie, element_term=term, pipeline=pipeline.name)
                    if term2 not in ix.groupings:
                        ix.groupings[term2] = LexicalGrouping(term=term2)
                    ix.groupings[term2].relationships.append(rel)
    return ix

def save_lexical_index(lexical_index: LexicalIndex, path: str):
    """
    Saves a YAML using standard mapping of datanodel to YAML

    :param lexical_index:
    :param path:
    :return:
    """
    yaml_dumper.dump(lexical_index, to_file=path)

def load_lexical_index(path: str) -> LexicalIndex:
    """
    Loads from a YAML file

    :param path:
    :return:
    """
    return yaml_loader.load(path, target_class=LexicalIndex)

def lexical_index_to_sssom(oi: BasicOntologyInterface, lexical_index: LexicalIndex, id='default',
                           ruleset: MappingRuleCollection = None) -> MappingSetDataFrame:
    """
    Transform a lexical index to an SSSOM MappingSetDataFrame by finding all pairs for any given index term

    :param oi:
    :param lexical_index:
    :param id:
    :return:
    """
    mappings = []
    logging.info('Converting lexical index to SSSOM')
    for term, grouping in lexical_index.groupings.items():
        elements = set([r.element for r in grouping.relationships])
        elementmap = defaultdict(list)
        for r in grouping.relationships:
            elementmap[r.element].append(r)
        if len(elementmap.keys()) < 2:
            continue
        for e1 in elementmap:
            for e2 in elementmap:
                if e1 < e2:
                    for r1 in elementmap[e1]:
                        for r2 in elementmap[e2]:
                            mappings.append(inferred_mapping(oi, term, r1, r2, ruleset=ruleset))

        #for r1 in grouping.relationships:
        #    for r2 in grouping.relationships:
        #        if r1.element < r2.element:
        #            mappings.append(create_mapping(oi, term, r1, r2))
    logging.info('Done creating SSSOM mappings')
    mset = MappingSet(mapping_set_id=id, mappings=mappings, license='CC-0')
    #doc = MappingSetDocument(prefix_map=oi.get_prefix_map(), mapping_set=mset)
    doc = MappingSetDocument(prefix_map={}, mapping_set=mset)
    return to_mapping_set_dataframe(doc)


def create_mapping(term: str, r1: RelationshipToTerm, r2: RelationshipToTerm,
                   pred: PRED_CURIE = SKOS_CLOSE_MATCH, confidence: float = None) -> Mapping:
    return Mapping(subject_id=r1.element,
                   object_id=r2.element,
                   predicate_id=pred,
                   confidence=confidence,
                   match_string=term,
                   subject_match_field=[r1.predicate],
                   object_match_field=[r2.predicate],
                   match_type=MatchTypeEnum.Lexical,
                   mapping_tool='oaklib'
                   )

def inferred_mapping(oi: BasicOntologyInterface, term: str, r1: RelationshipToTerm, r2: RelationshipToTerm,
                     ruleset: MappingRuleCollection = None) -> Mapping:
    m1 = create_mapping(term, r1, r2)
    m2 = create_mapping(term, r2, r1)
    weightmap: Dict[PRED_CURIE, float] = {}
    best: Tuple[float, Mapping, PRED_CURIE] = None, m1, m1.predicate_id
    if ruleset is not None:
        rules = ruleset.rules
    else:
        rules = []
    for rule in rules:
        inverted = False
        if precondition_holds(rule.preconditions, m1):
            m = m1
        elif not rule.oneway and precondition_holds(rule.preconditions, m2):
            m = m2
            inverted = True
        else:
            m = None
        if m:
            weight = 0.0
            if rule.postconditions.predicate_id:
                m.predicate_id = rule.postconditions.predicate_id
            if rule.postconditions.weight:
                weight = rule.postconditions.weight
            if inverted:
                inv_pred = invert_mapping_predicate(m.predicate_id)
                if inv_pred:
                    m.predicate_id = inv_pred
                else:
                    m = None
            if m:
                if m.predicate_id not in weightmap:
                    weightmap[m.predicate_id] = weight
                else:
                    weightmap[m.predicate_id] += weight
                weight = weightmap[m.predicate_id]
                if best[0] is None or weight > best[0]:
                    best = weight, m, m.predicate_id
                    #print(f' ** BEST {best} ==> {m}')
    best_weight, best_mapping, _ = best
    if best_weight is not None:
        best_mapping.confidence = inverse_logit(best_weight)
    best_mapping.subject_label = oi.get_label_by_curie(best_mapping.subject_id)
    best_mapping.object_label = oi.get_label_by_curie(best_mapping.object_id)
    return best_mapping

def inverse_logit(weight: float) -> float:
    """
    Inverse logit

    https://upload.wikimedia.org/wikipedia/commons/5/57/Logit.png

    :param weight:
    :return: probability between 0 and 1
    """
    return 1/(1+2**(-weight))

def invert_mapping_predicate(pred: PRED_CURIE) -> Optional[PRED_CURIE]:
    if pred == SKOS_EXACT_MATCH or pred == SKOS_CLOSE_MATCH:
        return pred
    if pred == SKOS_BROAD_MATCH:
        return SKOS_NARROW_MATCH
    return None

def precondition_holds(precondition: Precondition, mapping: Mapping) -> bool:
    for k in ['subject_match_field', 'object_match_field']:
        k_one_of = f'{k}_one_of'
        expected_values = getattr(precondition, k_one_of, [])
        actual_values = getattr(mapping, k, [])
        #print(f'CHECKING {actual_values} << {expected_values}')
        if expected_values and not any(v for v in actual_values if v in expected_values ):
            #print('** NO MATCH')
            return False
        #print(f'**** MTCH {k}')
    return True

def apply_transformation(term: str, transformation: LexicalTransformation) -> str:
    """
    Apply an individual transformation on a term

    :param term:
    :param transformation:
    :return:
    """
    typ = str(transformation.type)
    if typ == TransformationType.CaseNormalization.text:
        return term.lower()
    elif typ == TransformationType.WhitespaceNormalization.text:
        return re.sub(" {2,}", " ", term.strip())
    else:
        raise NotImplementedError(f'Transformation Type {typ} {type(typ)} not implemented {TransformationType.CaseNormalization.text}')

    
def save_mapping_rules(mapping_rules: MappingRuleCollection, path: str):
    """
    Saves a YAML using standard mapping of datanodel to YAML

    :param mapping_rules:
    :param path:
    :return:
    """
    yaml_dumper.dump(mapping_rules, to_file=path)

def load_mapping_rules(path: str) -> MappingRuleCollection:
    """
    Loads from a YAML file

    :param path:
    :return:
    """
    return yaml_loader.load(path, target_class=MappingRuleCollection)



from oaklib.datamodels.ontology_metadata import slots as omd_slots
import oaklib.datamodels.ontology_metadata as omd

NAMESPACES = [omd.SKOS, omd.RDF, omd.RDFS, omd.OIO]
DEFAULT_PREFIX_MAP = {ns.prefix: str(ns) for ns in NAMESPACES}

APP_NAME = 'ontology-access-kit'

# TODO: replace with oio vocab
LABEL_PREDICATE = omd.slots.label.curie
HAS_EXACT_SYNONYM = omd.slots.has_exact_synonym.curie
HAS_BROAD_SYNONYM = omd.slots.has_broad_synonym.curie
HAS_NARROW_SYNONYM = omd.slots.has_narrow_synonym.curie
HAS_RELATED_SYNONYM = omd.slots.has_related_synonym.curie
SYNONYM_PREDICATES = [HAS_RELATED_SYNONYM, HAS_NARROW_SYNONYM, HAS_EXACT_SYNONYM, HAS_BROAD_SYNONYM]

SCOPE_TO_SYNONYM_PRED_MAP = {
    'EXACT': HAS_EXACT_SYNONYM,
    'RELATED': HAS_RELATED_SYNONYM,
    'NARROW': HAS_NARROW_SYNONYM,
    'BROAD': HAS_BROAD_SYNONYM
}

SUBCLASS_OF = omd.slots.subClassOf.curie
IS_A = omd.slots.subClassOf.curie
EQUIVALENT_CLASS = 'owl:equivalentClass'
OWL_SAME_AS = 'owl:sameAs'
PART_OF = 'BFO:0000050'
HAS_PART = 'BFO:0000051'
ONLY_IN_TAXON = 'RO:0002160'
NEVER_IN_TAXON = 'RO:0002161'
IN_TAXON = 'RO:0002162'
PRESENT_IN_TAXON = 'RO:0002175'

OBO_PURL = 'http://purl.obolibrary.org/obo/'
BIOPORTAL_PURL = 'http://purl.bioontology.org/ontology/'

IN_SUBSET = omd.slots.in_subset.curie

SKOS_EXACT_MATCH = omd.slots.exactMatch.curie
SKOS_CLOSE_MATCH = omd.slots.closeMatch.curie
SKOS_NARROW_MATCH = omd.slots.narrowMatch.curie
SKOS_BROAD_MATCH = omd.slots.broadMatch.curie
#SKOS_RELATED_MATCH = omd.slots.relatedMatch.curie
SKOS_RELATED_MATCH = 'skos:relatedMatch'
SKOS_MATCH_PREDICATES = [SKOS_BROAD_MATCH, SKOS_NARROW_MATCH, SKOS_CLOSE_MATCH, SKOS_EXACT_MATCH, SKOS_RELATED_MATCH]
HAS_DBXREF = omd.slots.database_cross_reference.curie
ALL_MATCH_PREDICATES = SKOS_MATCH_PREDICATES + [HAS_DBXREF, OWL_SAME_AS]
HAS_DEFINITION_URI = omd.slots.definition.uri


# Auto generated from taxon_constraints.yaml by pythongen.py version: 0.9.0
# Generation date: 2022-04-17T13:37:35
# Schema: taxon-constraints
#
# id: https://w3id.org/linkml/taxon_constraints
# description: A datamodel for representing inferred and asserted taxon constraints
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import sys
import re
from jsonasobj2 import JsonObj, as_dict
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
from linkml_runtime.linkml_model.meta import EnumDefinition, PermissibleValue, PvFormulaOptions

from linkml_runtime.utils.slot import Slot
from linkml_runtime.utils.metamodelcore import empty_list, empty_dict, bnode
from linkml_runtime.utils.yamlutils import YAMLRoot, extended_str, extended_float, extended_int
from linkml_runtime.utils.dataclass_extensions_376 import dataclasses_init_fn_with_kwargs
from linkml_runtime.utils.formatutils import camelcase, underscore, sfx
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from rdflib import Namespace, URIRef
from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_runtime.linkml_model.types import Boolean, String, Uriorcurie
from linkml_runtime.utils.metamodelcore import Bool, URIorCURIE

metamodel_version = "1.7.0"
version = None

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
NCBITAXON = CurieNamespace('NCBITaxon', 'http://example.org/UNKNOWN/NCBITaxon/')
NCBITAXON_UNION = CurieNamespace('NCBITaxon_Union', 'http://example.org/UNKNOWN/NCBITaxon_Union/')
RO = CurieNamespace('RO', 'http://purl.obolibrary.org/obo/RO_')
LINKML = CurieNamespace('linkml', 'https://w3id.org/linkml/')
OWL = CurieNamespace('owl', 'http://www.w3.org/2002/07/owl#')
PAV = CurieNamespace('pav', 'http://purl.org/pav/')
PROV = CurieNamespace('prov', 'http://www.w3.org/ns/prov#')
RDF = CurieNamespace('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#')
RDFS = CurieNamespace('rdfs', 'http://www.w3.org/2000/01/rdf-schema#')
SCHEMA = CurieNamespace('schema', 'http://schema.org/')
SH = CurieNamespace('sh', 'https://w3id.org/shacl/')
SKOS = CurieNamespace('skos', 'http://www.w3.org/2004/02/skos/core#')
TC = CurieNamespace('tc', 'https://w3id.org/linkml/taxon_constraints/')
XSD = CurieNamespace('xsd', 'http://www.w3.org/2001/XMLSchema#')
DEFAULT_ = TC


# Types

# Class references
class TermId(URIorCURIE):
    pass


class SubjectTermId(TermId):
    pass


class TaxonId(TermId):
    pass


class PredicateTermId(TermId):
    pass


@dataclass
class Term(YAMLRoot):
    """
    An ontology term. In this model this is either the SubjectTerm of a taxon constraint, or an actual taxon
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OWL.Class
    class_class_curie: ClassVar[str] = "owl:Class"
    class_name: ClassVar[str] = "Term"
    class_model_uri: ClassVar[URIRef] = TC.Term

    id: Union[str, TermId] = None
    label: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.id):
            self.MissingRequiredField("id")
        if not isinstance(self.id, TermId):
            self.id = TermId(self.id)

        if self.label is not None and not isinstance(self.label, str):
            self.label = str(self.label)

        super().__post_init__(**kwargs)


@dataclass
class SubjectTerm(Term):
    """
    A term that is the subject of a taxon constraint. Typically comes from ontologies like GO, UBERON, CL, ...
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = TC.SubjectTerm
    class_class_curie: ClassVar[str] = "tc:SubjectTerm"
    class_name: ClassVar[str] = "SubjectTerm"
    class_model_uri: ClassVar[URIRef] = TC.SubjectTerm

    id: Union[str, SubjectTermId] = None
    description: Optional[str] = None
    unsatisfiable: Optional[Union[bool, Bool]] = None
    only_in: Optional[Union[Union[dict, "TaxonConstraint"], List[Union[dict, "TaxonConstraint"]]]] = empty_list()
    never_in: Optional[Union[Union[dict, "TaxonConstraint"], List[Union[dict, "TaxonConstraint"]]]] = empty_list()
    present_in: Optional[Union[Union[dict, "TaxonConstraint"], List[Union[dict, "TaxonConstraint"]]]] = empty_list()
    present_in_ancestor_of: Optional[Union[Union[dict, "TaxonConstraint"], List[Union[dict, "TaxonConstraint"]]]] = empty_list()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.id):
            self.MissingRequiredField("id")
        if not isinstance(self.id, SubjectTermId):
            self.id = SubjectTermId(self.id)

        if self.description is not None and not isinstance(self.description, str):
            self.description = str(self.description)

        if self.unsatisfiable is not None and not isinstance(self.unsatisfiable, Bool):
            self.unsatisfiable = Bool(self.unsatisfiable)

        if not isinstance(self.only_in, list):
            self.only_in = [self.only_in] if self.only_in is not None else []
        self.only_in = [v if isinstance(v, TaxonConstraint) else TaxonConstraint(**as_dict(v)) for v in self.only_in]

        if not isinstance(self.never_in, list):
            self.never_in = [self.never_in] if self.never_in is not None else []
        self.never_in = [v if isinstance(v, TaxonConstraint) else TaxonConstraint(**as_dict(v)) for v in self.never_in]

        if not isinstance(self.present_in, list):
            self.present_in = [self.present_in] if self.present_in is not None else []
        self.present_in = [v if isinstance(v, TaxonConstraint) else TaxonConstraint(**as_dict(v)) for v in self.present_in]

        if not isinstance(self.present_in_ancestor_of, list):
            self.present_in_ancestor_of = [self.present_in_ancestor_of] if self.present_in_ancestor_of is not None else []
        self.present_in_ancestor_of = [v if isinstance(v, TaxonConstraint) else TaxonConstraint(**as_dict(v)) for v in self.present_in_ancestor_of]

        super().__post_init__(**kwargs)


@dataclass
class Taxon(Term):
    """
    A term that represents a taxonomic group
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = TC.Taxon
    class_class_curie: ClassVar[str] = "tc:Taxon"
    class_name: ClassVar[str] = "Taxon"
    class_model_uri: ClassVar[URIRef] = TC.Taxon

    id: Union[str, TaxonId] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.id):
            self.MissingRequiredField("id")
        if not isinstance(self.id, TaxonId):
            self.id = TaxonId(self.id)

        super().__post_init__(**kwargs)


@dataclass
class PredicateTerm(Term):
    """
    A term that represents a relationship type
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = TC.PredicateTerm
    class_class_curie: ClassVar[str] = "tc:PredicateTerm"
    class_name: ClassVar[str] = "PredicateTerm"
    class_model_uri: ClassVar[URIRef] = TC.PredicateTerm

    id: Union[str, PredicateTermId] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.id):
            self.MissingRequiredField("id")
        if not isinstance(self.id, PredicateTermId):
            self.id = PredicateTermId(self.id)

        super().__post_init__(**kwargs)


@dataclass
class TaxonConstraint(YAMLRoot):
    """
    An individual taxon constraint
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = RDF.Statement
    class_class_curie: ClassVar[str] = "rdf:Statement"
    class_name: ClassVar[str] = "TaxonConstraint"
    class_model_uri: ClassVar[URIRef] = TC.TaxonConstraint

    subject: Optional[Union[str, SubjectTermId]] = None
    predicate: Optional[Union[str, PredicateTermId]] = None
    asserted: Optional[Union[bool, Bool]] = None
    evolutionary: Optional[Union[bool, Bool]] = None
    redundant: Optional[Union[bool, Bool]] = None
    redundant_with_only_in: Optional[Union[bool, Bool]] = None
    taxon: Optional[Union[dict, Taxon]] = None
    redundant_with: Optional[Union[Union[dict, "TaxonConstraint"], List[Union[dict, "TaxonConstraint"]]]] = empty_list()
    contradicted_by: Optional[Union[Union[dict, "TaxonConstraint"], List[Union[dict, "TaxonConstraint"]]]] = empty_list()
    via_terms: Optional[Union[Dict[Union[str, SubjectTermId], Union[dict, SubjectTerm]], List[Union[dict, SubjectTerm]]]] = empty_dict()
    predicates: Optional[Union[Union[str, PredicateTermId], List[Union[str, PredicateTermId]]]] = empty_list()
    sources: Optional[Union[Union[str, URIorCURIE], List[Union[str, URIorCURIE]]]] = empty_list()
    comments: Optional[Union[str, List[str]]] = empty_list()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.subject is not None and not isinstance(self.subject, SubjectTermId):
            self.subject = SubjectTermId(self.subject)

        if self.predicate is not None and not isinstance(self.predicate, PredicateTermId):
            self.predicate = PredicateTermId(self.predicate)

        if self.asserted is not None and not isinstance(self.asserted, Bool):
            self.asserted = Bool(self.asserted)

        if self.evolutionary is not None and not isinstance(self.evolutionary, Bool):
            self.evolutionary = Bool(self.evolutionary)

        if self.redundant is not None and not isinstance(self.redundant, Bool):
            self.redundant = Bool(self.redundant)

        if self.redundant_with_only_in is not None and not isinstance(self.redundant_with_only_in, Bool):
            self.redundant_with_only_in = Bool(self.redundant_with_only_in)

        if self.taxon is not None and not isinstance(self.taxon, Taxon):
            self.taxon = Taxon(**as_dict(self.taxon))

        if not isinstance(self.redundant_with, list):
            self.redundant_with = [self.redundant_with] if self.redundant_with is not None else []
        self.redundant_with = [v if isinstance(v, TaxonConstraint) else TaxonConstraint(**as_dict(v)) for v in self.redundant_with]

        if not isinstance(self.contradicted_by, list):
            self.contradicted_by = [self.contradicted_by] if self.contradicted_by is not None else []
        self.contradicted_by = [v if isinstance(v, TaxonConstraint) else TaxonConstraint(**as_dict(v)) for v in self.contradicted_by]

        self._normalize_inlined_as_list(slot_name="via_terms", slot_type=SubjectTerm, key_name="id", keyed=True)

        if not isinstance(self.predicates, list):
            self.predicates = [self.predicates] if self.predicates is not None else []
        self.predicates = [v if isinstance(v, PredicateTermId) else PredicateTermId(v) for v in self.predicates]

        if not isinstance(self.sources, list):
            self.sources = [self.sources] if self.sources is not None else []
        self.sources = [v if isinstance(v, URIorCURIE) else URIorCURIE(v) for v in self.sources]

        if not isinstance(self.comments, list):
            self.comments = [self.comments] if self.comments is not None else []
        self.comments = [v if isinstance(v, str) else str(v) for v in self.comments]

        super().__post_init__(**kwargs)


# Enumerations


# Slots
class slots:
    pass

slots.term__id = Slot(uri=TC.id, name="term__id", curie=TC.curie('id'),
                   model_uri=TC.term__id, domain=None, range=URIRef)

slots.term__label = Slot(uri=RDFS.label, name="term__label", curie=RDFS.curie('label'),
                   model_uri=TC.term__label, domain=None, range=Optional[str])

slots.subjectTerm__description = Slot(uri=TC.description, name="subjectTerm__description", curie=TC.curie('description'),
                   model_uri=TC.subjectTerm__description, domain=None, range=Optional[str])

slots.subjectTerm__unsatisfiable = Slot(uri=TC.unsatisfiable, name="subjectTerm__unsatisfiable", curie=TC.curie('unsatisfiable'),
                   model_uri=TC.subjectTerm__unsatisfiable, domain=None, range=Optional[Union[bool, Bool]])

slots.subjectTerm__only_in = Slot(uri=RO['0002160'], name="subjectTerm__only_in", curie=RO.curie('0002160'),
                   model_uri=TC.subjectTerm__only_in, domain=None, range=Optional[Union[Union[dict, TaxonConstraint], List[Union[dict, TaxonConstraint]]]])

slots.subjectTerm__never_in = Slot(uri=RO['0002161'], name="subjectTerm__never_in", curie=RO.curie('0002161'),
                   model_uri=TC.subjectTerm__never_in, domain=None, range=Optional[Union[Union[dict, TaxonConstraint], List[Union[dict, TaxonConstraint]]]])

slots.subjectTerm__present_in = Slot(uri=RO['0002175'], name="subjectTerm__present_in", curie=RO.curie('0002175'),
                   model_uri=TC.subjectTerm__present_in, domain=None, range=Optional[Union[Union[dict, TaxonConstraint], List[Union[dict, TaxonConstraint]]]])

slots.subjectTerm__present_in_ancestor_of = Slot(uri=TC.present_in_ancestor_of, name="subjectTerm__present_in_ancestor_of", curie=TC.curie('present_in_ancestor_of'),
                   model_uri=TC.subjectTerm__present_in_ancestor_of, domain=None, range=Optional[Union[Union[dict, TaxonConstraint], List[Union[dict, TaxonConstraint]]]])

slots.taxonConstraint__subject = Slot(uri=RDF.subject, name="taxonConstraint__subject", curie=RDF.curie('subject'),
                   model_uri=TC.taxonConstraint__subject, domain=None, range=Optional[Union[str, SubjectTermId]])

slots.taxonConstraint__predicate = Slot(uri=RDF.predicate, name="taxonConstraint__predicate", curie=RDF.curie('predicate'),
                   model_uri=TC.taxonConstraint__predicate, domain=None, range=Optional[Union[str, PredicateTermId]])

slots.taxonConstraint__asserted = Slot(uri=TC.asserted, name="taxonConstraint__asserted", curie=TC.curie('asserted'),
                   model_uri=TC.taxonConstraint__asserted, domain=None, range=Optional[Union[bool, Bool]])

slots.taxonConstraint__evolutionary = Slot(uri=TC.evolutionary, name="taxonConstraint__evolutionary", curie=TC.curie('evolutionary'),
                   model_uri=TC.taxonConstraint__evolutionary, domain=None, range=Optional[Union[bool, Bool]])

slots.taxonConstraint__redundant = Slot(uri=TC.redundant, name="taxonConstraint__redundant", curie=TC.curie('redundant'),
                   model_uri=TC.taxonConstraint__redundant, domain=None, range=Optional[Union[bool, Bool]])

slots.taxonConstraint__redundant_with_only_in = Slot(uri=TC.redundant_with_only_in, name="taxonConstraint__redundant_with_only_in", curie=TC.curie('redundant_with_only_in'),
                   model_uri=TC.taxonConstraint__redundant_with_only_in, domain=None, range=Optional[Union[bool, Bool]])

slots.taxonConstraint__taxon = Slot(uri=RDF.object, name="taxonConstraint__taxon", curie=RDF.curie('object'),
                   model_uri=TC.taxonConstraint__taxon, domain=None, range=Optional[Union[dict, Taxon]])

slots.taxonConstraint__redundant_with = Slot(uri=TC.redundant_with, name="taxonConstraint__redundant_with", curie=TC.curie('redundant_with'),
                   model_uri=TC.taxonConstraint__redundant_with, domain=None, range=Optional[Union[Union[dict, TaxonConstraint], List[Union[dict, TaxonConstraint]]]])

slots.taxonConstraint__contradicted_by = Slot(uri=TC.contradicted_by, name="taxonConstraint__contradicted_by", curie=TC.curie('contradicted_by'),
                   model_uri=TC.taxonConstraint__contradicted_by, domain=None, range=Optional[Union[Union[dict, TaxonConstraint], List[Union[dict, TaxonConstraint]]]])

slots.taxonConstraint__via_terms = Slot(uri=TC.via_terms, name="taxonConstraint__via_terms", curie=TC.curie('via_terms'),
                   model_uri=TC.taxonConstraint__via_terms, domain=None, range=Optional[Union[Dict[Union[str, SubjectTermId], Union[dict, SubjectTerm]], List[Union[dict, SubjectTerm]]]])

slots.taxonConstraint__predicates = Slot(uri=TC.predicates, name="taxonConstraint__predicates", curie=TC.curie('predicates'),
                   model_uri=TC.taxonConstraint__predicates, domain=None, range=Optional[Union[Union[str, PredicateTermId], List[Union[str, PredicateTermId]]]])

slots.taxonConstraint__sources = Slot(uri=TC.sources, name="taxonConstraint__sources", curie=TC.curie('sources'),
                   model_uri=TC.taxonConstraint__sources, domain=None, range=Optional[Union[Union[str, URIorCURIE], List[Union[str, URIorCURIE]]]])

slots.taxonConstraint__comments = Slot(uri=TC.comments, name="taxonConstraint__comments", curie=TC.curie('comments'),
                   model_uri=TC.taxonConstraint__comments, domain=None, range=Optional[Union[str, List[str]]])



# Auto generated from text_annotator.yaml by pythongen.py version: 0.9.0
# Generation date: 2022-04-05T17:21:12
# Schema: text-annotator
#
# id: https://w3id.org/linkml/text_annotator
# description: A datamodel for representing the results of textual named entity recognition annotation results
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import sys
import re
from jsonasobj2 import JsonObj, as_dict
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
from linkml_runtime.linkml_model.meta import EnumDefinition, PermissibleValue, PvFormulaOptions

from linkml_runtime.utils.slot import Slot
from linkml_runtime.utils.metamodelcore import empty_list, empty_dict, bnode
from linkml_runtime.utils.yamlutils import YAMLRoot, extended_str, extended_float, extended_int
from linkml_runtime.utils.dataclass_extensions_376 import dataclasses_init_fn_with_kwargs
from linkml_runtime.utils.formatutils import camelcase, underscore, sfx
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from rdflib import Namespace, URIRef
from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_runtime.linkml_model.types import Boolean, Float, Integer, String, Uriorcurie
from linkml_runtime.utils.metamodelcore import Bool, URIorCURIE

metamodel_version = "1.7.0"
version = None

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
ANN = CurieNamespace('ann', 'https://w3id.org/linkml/text_annotator/')
BPA = CurieNamespace('bpa', 'https://bioportal.bioontology.org/annotator/')
LINKML = CurieNamespace('linkml', 'https://w3id.org/linkml/')
OWL = CurieNamespace('owl', 'http://www.w3.org/2002/07/owl#')
PAV = CurieNamespace('pav', 'http://purl.org/pav/')
PROV = CurieNamespace('prov', 'http://www.w3.org/ns/prov#')
RDF = CurieNamespace('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#')
RDFS = CurieNamespace('rdfs', 'http://www.w3.org/2000/01/rdf-schema#')
SCHEMA = CurieNamespace('schema', 'http://schema.org/')
SH = CurieNamespace('sh', 'https://w3id.org/shacl/')
SKOS = CurieNamespace('skos', 'http://www.w3.org/2004/02/skos/core#')
SSSOM = CurieNamespace('sssom', 'http://w3id.org/sssom/')
XSD = CurieNamespace('xsd', 'http://www.w3.org/2001/XMLSchema#')
DEFAULT_ = ANN


# Types
class Position(Integer):
    type_class_uri = XSD.integer
    type_class_curie = "xsd:integer"
    type_name = "Position"
    type_model_uri = ANN.Position


# Class references
class TextualElementId(URIorCURIE):
    pass


@dataclass
class TextAnnotationResultSet(YAMLRoot):
    """
    A collection of annotation results
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = ANN.TextAnnotationResultSet
    class_class_curie: ClassVar[str] = "ann:TextAnnotationResultSet"
    class_name: ClassVar[str] = "TextAnnotationResultSet"
    class_model_uri: ClassVar[URIRef] = ANN.TextAnnotationResultSet

    annotations: Optional[Union[Union[dict, "TextAnnotation"], List[Union[dict, "TextAnnotation"]]]] = empty_list()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if not isinstance(self.annotations, list):
            self.annotations = [self.annotations] if self.annotations is not None else []
        self.annotations = [v if isinstance(v, TextAnnotation) else TextAnnotation(**as_dict(v)) for v in self.annotations]

        super().__post_init__(**kwargs)


@dataclass
class TextualElement(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = ANN.TextualElement
    class_class_curie: ClassVar[str] = "ann:TextualElement"
    class_name: ClassVar[str] = "TextualElement"
    class_model_uri: ClassVar[URIRef] = ANN.TextualElement

    id: Union[str, TextualElementId] = None
    text: Optional[str] = None
    source_text: Optional[str] = None
    parent_document: Optional[Union[str, URIorCURIE]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.id):
            self.MissingRequiredField("id")
        if not isinstance(self.id, TextualElementId):
            self.id = TextualElementId(self.id)

        if self.text is not None and not isinstance(self.text, str):
            self.text = str(self.text)

        if self.source_text is not None and not isinstance(self.source_text, str):
            self.source_text = str(self.source_text)

        if self.parent_document is not None and not isinstance(self.parent_document, URIorCURIE):
            self.parent_document = URIorCURIE(self.parent_document)

        super().__post_init__(**kwargs)


@dataclass
class HasSpan(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = ANN.HasSpan
    class_class_curie: ClassVar[str] = "ann:HasSpan"
    class_name: ClassVar[str] = "HasSpan"
    class_model_uri: ClassVar[URIRef] = ANN.HasSpan

    subject_start: Optional[Union[int, Position]] = None
    subject_end: Optional[Union[int, Position]] = None
    subject_label: Optional[str] = None
    subject_source: Optional[str] = None
    subject_text_id: Optional[Union[str, TextualElementId]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.subject_start is not None and not isinstance(self.subject_start, Position):
            self.subject_start = Position(self.subject_start)

        if self.subject_end is not None and not isinstance(self.subject_end, Position):
            self.subject_end = Position(self.subject_end)

        if self.subject_label is not None and not isinstance(self.subject_label, str):
            self.subject_label = str(self.subject_label)

        if self.subject_source is not None and not isinstance(self.subject_source, str):
            self.subject_source = str(self.subject_source)

        if self.subject_text_id is not None and not isinstance(self.subject_text_id, TextualElementId):
            self.subject_text_id = TextualElementId(self.subject_text_id)

        super().__post_init__(**kwargs)


@dataclass
class TextAnnotation(YAMLRoot):
    """
    An individual text annotation
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = ANN.TextAnnotation
    class_class_curie: ClassVar[str] = "ann:TextAnnotation"
    class_name: ClassVar[str] = "TextAnnotation"
    class_model_uri: ClassVar[URIRef] = ANN.TextAnnotation

    predicate_id: Optional[str] = None
    object_id: Optional[str] = None
    object_label: Optional[str] = None
    object_source: Optional[str] = None
    confidence: Optional[float] = None
    match_string: Optional[str] = None
    is_longest_match: Optional[Union[bool, Bool]] = None
    match_type: Optional[str] = None
    info: Optional[str] = None
    subject_start: Optional[Union[int, Position]] = None
    subject_end: Optional[Union[int, Position]] = None
    subject_label: Optional[str] = None
    subject_source: Optional[str] = None
    subject_text_id: Optional[Union[str, TextualElementId]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.predicate_id is not None and not isinstance(self.predicate_id, str):
            self.predicate_id = str(self.predicate_id)

        if self.object_id is not None and not isinstance(self.object_id, str):
            self.object_id = str(self.object_id)

        if self.object_label is not None and not isinstance(self.object_label, str):
            self.object_label = str(self.object_label)

        if self.object_source is not None and not isinstance(self.object_source, str):
            self.object_source = str(self.object_source)

        if self.confidence is not None and not isinstance(self.confidence, float):
            self.confidence = float(self.confidence)

        if self.match_string is not None and not isinstance(self.match_string, str):
            self.match_string = str(self.match_string)

        if self.is_longest_match is not None and not isinstance(self.is_longest_match, Bool):
            self.is_longest_match = Bool(self.is_longest_match)

        if self.match_type is not None and not isinstance(self.match_type, str):
            self.match_type = str(self.match_type)

        if self.info is not None and not isinstance(self.info, str):
            self.info = str(self.info)

        if self.subject_start is not None and not isinstance(self.subject_start, Position):
            self.subject_start = Position(self.subject_start)

        if self.subject_end is not None and not isinstance(self.subject_end, Position):
            self.subject_end = Position(self.subject_end)

        if self.subject_label is not None and not isinstance(self.subject_label, str):
            self.subject_label = str(self.subject_label)

        if self.subject_source is not None and not isinstance(self.subject_source, str):
            self.subject_source = str(self.subject_source)

        if self.subject_text_id is not None and not isinstance(self.subject_text_id, TextualElementId):
            self.subject_text_id = TextualElementId(self.subject_text_id)

        super().__post_init__(**kwargs)


# Enumerations
class TransformationType(EnumDefinitionImpl):
    """
    A controlled datamodels of the types of transformation that can be applied to
    """
    Stemming = PermissibleValue(text="Stemming",
                                       description="Removal of the last few characters of a word to yield a stem term for each word in the term")
    Lemmatization = PermissibleValue(text="Lemmatization",
                                                 description="Contextual reduction of a word to its base form for each word in the term")
    WordOrderNormalization = PermissibleValue(text="WordOrderNormalization",
                                                                   description="reorder words in the term to a standard order such that comparisons are order-independent")
    Depluralization = PermissibleValue(text="Depluralization",
                                                     description="Transform plural form to singular form for each word in a term")
    CaseNormalization = PermissibleValue(text="CaseNormalization",
                                                         description="Transform term to a standard case, typically lowercase")
    WhitespaceNormalization = PermissibleValue(text="WhitespaceNormalization",
                                                                     description="Trim whitespace, condense whitespace runs, and transform all non-space whitespace to spaces")
    TermExpanson = PermissibleValue(text="TermExpanson",
                                               description="Expand terms using a dictionary")

    _defn = EnumDefinition(
        name="TransformationType",
        description="A controlled datamodels of the types of transformation that can be applied to",
    )

# Slots
class slots:
    pass

slots.textAnnotationResultSet__annotations = Slot(uri=ANN.annotations, name="textAnnotationResultSet__annotations", curie=ANN.curie('annotations'),
                   model_uri=ANN.textAnnotationResultSet__annotations, domain=None, range=Optional[Union[Union[dict, TextAnnotation], List[Union[dict, TextAnnotation]]]])

slots.textualElement__id = Slot(uri=ANN.id, name="textualElement__id", curie=ANN.curie('id'),
                   model_uri=ANN.textualElement__id, domain=None, range=URIRef)

slots.textualElement__text = Slot(uri=ANN.text, name="textualElement__text", curie=ANN.curie('text'),
                   model_uri=ANN.textualElement__text, domain=None, range=Optional[str])

slots.textualElement__source_text = Slot(uri=ANN.source_text, name="textualElement__source_text", curie=ANN.curie('source_text'),
                   model_uri=ANN.textualElement__source_text, domain=None, range=Optional[str])

slots.textualElement__parent_document = Slot(uri=ANN.parent_document, name="textualElement__parent_document", curie=ANN.curie('parent_document'),
                   model_uri=ANN.textualElement__parent_document, domain=None, range=Optional[Union[str, URIorCURIE]])

slots.hasSpan__subject_start = Slot(uri=ANN.subject_start, name="hasSpan__subject_start", curie=ANN.curie('subject_start'),
                   model_uri=ANN.hasSpan__subject_start, domain=None, range=Optional[Union[int, Position]])

slots.hasSpan__subject_end = Slot(uri=ANN.subject_end, name="hasSpan__subject_end", curie=ANN.curie('subject_end'),
                   model_uri=ANN.hasSpan__subject_end, domain=None, range=Optional[Union[int, Position]])

slots.hasSpan__subject_label = Slot(uri=ANN.subject_label, name="hasSpan__subject_label", curie=ANN.curie('subject_label'),
                   model_uri=ANN.hasSpan__subject_label, domain=None, range=Optional[str])

slots.hasSpan__subject_source = Slot(uri=SSSOM.subject_source, name="hasSpan__subject_source", curie=SSSOM.curie('subject_source'),
                   model_uri=ANN.hasSpan__subject_source, domain=None, range=Optional[str])

slots.hasSpan__subject_text_id = Slot(uri=ANN.subject_text_id, name="hasSpan__subject_text_id", curie=ANN.curie('subject_text_id'),
                   model_uri=ANN.hasSpan__subject_text_id, domain=None, range=Optional[Union[str, TextualElementId]])

slots.textAnnotation__predicate_id = Slot(uri=SSSOM.predicate_id, name="textAnnotation__predicate_id", curie=SSSOM.curie('predicate_id'),
                   model_uri=ANN.textAnnotation__predicate_id, domain=None, range=Optional[str])

slots.textAnnotation__object_id = Slot(uri=SSSOM.object_id, name="textAnnotation__object_id", curie=SSSOM.curie('object_id'),
                   model_uri=ANN.textAnnotation__object_id, domain=None, range=Optional[str])

slots.textAnnotation__object_label = Slot(uri=SSSOM.object_label, name="textAnnotation__object_label", curie=SSSOM.curie('object_label'),
                   model_uri=ANN.textAnnotation__object_label, domain=None, range=Optional[str])

slots.textAnnotation__object_source = Slot(uri=SSSOM.object_source, name="textAnnotation__object_source", curie=SSSOM.curie('object_source'),
                   model_uri=ANN.textAnnotation__object_source, domain=None, range=Optional[str])

slots.textAnnotation__confidence = Slot(uri=SSSOM.confidence, name="textAnnotation__confidence", curie=SSSOM.curie('confidence'),
                   model_uri=ANN.textAnnotation__confidence, domain=None, range=Optional[float])

slots.textAnnotation__match_string = Slot(uri=SSSOM.match_string, name="textAnnotation__match_string", curie=SSSOM.curie('match_string'),
                   model_uri=ANN.textAnnotation__match_string, domain=None, range=Optional[str])

slots.textAnnotation__is_longest_match = Slot(uri=ANN.is_longest_match, name="textAnnotation__is_longest_match", curie=ANN.curie('is_longest_match'),
                   model_uri=ANN.textAnnotation__is_longest_match, domain=None, range=Optional[Union[bool, Bool]])

slots.textAnnotation__match_type = Slot(uri=ANN.match_type, name="textAnnotation__match_type", curie=ANN.curie('match_type'),
                   model_uri=ANN.textAnnotation__match_type, domain=None, range=Optional[str])

slots.textAnnotation__info = Slot(uri=ANN.info, name="textAnnotation__info", curie=ANN.curie('info'),
                   model_uri=ANN.textAnnotation__info, domain=None, range=Optional[str])
